#!/usr/bin/env python

# -*- coding: utf-8 -*-
"""
Created on Wed Apr 19 15:56:54 2019

@author: ke291

Contains all of the specific code for running Gaussian jobs on Ziggy.
A lot of code is reused from Gaussian.py. Called by PyDP4.py.
"""

import subprocess
import socket
import os
import time
import math

import dp5.old_files.Gaussian as Gaussian

SetupNMRCalcs = Gaussian.SetupNMRCalcs

SetupECalcs = Gaussian.SetupECalcs

SetupOptCalcs = Gaussian.SetupOptCalcs

ReadShieldings = Gaussian.ReadShieldings

ReadEnergies = Gaussian.ReadEnergies

ReadGeometries = Gaussian.ReadGeometries

IsGausCompleted = Gaussian.IsGausCompleted

Converged = Gaussian.Converged


def RunNMRCalcs(Isomers, settings):
    print('\nRunning Gaussian NMR calculations on Ziggy...')

    # Run Gaussian jobs on Ziggy cluster in folder named after date and time
    # Split in batches, if needed

    jobdir = os.getcwd()
    os.chdir('nmr')

    GausJobs = []

    for iso in Isomers:
        GausJobs.extend([x for x in iso.NMRInputFiles if (
            x[:-4] + '.out') not in iso.NMROutputFiles])

    Completed = RunCalcs(GausJobs, settings)

    for iso in Isomers:
        iso.NMROutputFiles.extend(
            [x[:-4] + '.out' for x in iso.NMRInputFiles if (x[:-4] + '.out') in Completed])

    os.chdir(jobdir)

    return Isomers


def RunECalcs(Isomers, settings):
    print('\nRunning Gaussian DFT energy calculations on Ziggy...')

    jobdir = os.getcwd()
    os.chdir('e')

    GausJobs = []

    for iso in Isomers:
        GausJobs.extend([x for x in iso.EInputFiles if (
            x[:-4] + '.out') not in iso.EOutputFiles])

    Completed = RunCalcs(GausJobs, settings)

    for iso in Isomers:
        iso.EOutputFiles.extend(
            [x[:-4] + '.out' for x in iso.EInputFiles if (x[:-4] + '.out') in Completed])

    os.chdir(jobdir)

    return Isomers


def RunOptCalcs(Isomers, settings):
    print('\nRunning Gaussian DFT geometry optimizations on Ziggy...')

    jobdir = os.getcwd()
    os.chdir('opt')

    GausJobs = []

    for iso in Isomers:
        GausJobs.extend([x for x in iso.OptInputFiles if (
            x[:-4] + '.out') not in iso.OptOutputFiles])

    Completed = RunCalcs(GausJobs, settings)

    for iso in Isomers:
        iso.OptOutputFiles.extend(
            [x[:-4] + '.out' for x in iso.OptInputFiles if (x[:-4] + '.out') in Completed])

    os.chdir(jobdir)

    return Isomers


def RunCalcs(GausJobs, settings):

    MaxCon = settings.MaxConcurrentJobsZiggy

    if len(GausJobs) < MaxCon:
        if len(GausJobs) > 0:
            RunBatchOnZiggy(0, settings.queue, GausJobs, settings)
    else:
        if len(GausJobs) > 0:
            print("The DFT calculations will be done in " +
                  str(math.ceil(len(GausJobs) / MaxCon)) + " batches")
            i = 0
            while (i + 1) * MaxCon < len(GausJobs):
                print("Starting batch nr " + str(i + 1))
                RunBatchOnZiggy(str(i + 1), settings.queue,
                                GausJobs[(i * MaxCon):((i + 1) * MaxCon)], settings)
                i += 1
            print("Starting batch nr " + str(i + 1))
            RunBatchOnZiggy(str(i + 1), settings.queue,
                            GausJobs[(i * MaxCon):], settings)

    NCompleted = 0
    Completed = []

    for f in GausJobs:
        if IsGausCompleted(f[:-4] + '.out'):
            Completed.append(f[:-4] + '.out')
            NCompleted += 1

    if NCompleted > 0:
        print(str(NCompleted) + " Gaussian jobs of " + str(len(GausJobs)) +
              " completed successfully.")
    elif len(GausJobs) == 0:
        print("There were no jobs to run.")

    return Completed


def RunBatchOnZiggy(findex, queue, GausFiles, settings):

    # Run Gaussian jobs on Ziggy cluster in folder named after date and time
    # Split in batches, if needed

    if findex == 0:
        folder = settings.StartTime + settings.Title
    else:
        folder = settings.StartTime + findex + settings.Title

    print("Setting up jobs for running on ziggy...\n")

    # Check that folder does not exist, create job folder on ziggy
    outp = subprocess.check_output('ssh ziggy ls', shell=True)
    if folder in outp.decode():
        print("Folder exists on ziggy, choose another folder name.")
        return

    outp = subprocess.check_output('ssh ziggy mkdir ' + folder, shell=True)

    print("Results folder: " + folder)

    # Write the qsub scripts
    for f in GausFiles:
        WriteSubScript(f, queue, folder, settings)
    print(str(len(GausFiles)) + ' slurm scripts generated')

    # Upload .com files and .qsub files to directory
    print("Uploading files to ziggy...")
    for f in GausFiles:
        outp = subprocess.check_output('scp ' + f + ' ziggy:~/' + folder,
                                       shell=True)
        outp = subprocess.check_output('scp ' + f[:-4] + 'slurm ziggy:~/' +
                                       folder, shell=True)

    print(str(len(GausFiles)) + ' .com and slurm files uploaded to ziggy')

    JobIDs = []
    # Launch the calculations
    for f in GausFiles:
        job = '~/' + folder + '/' + f[:-4]
        outp = subprocess.check_output(
            'ssh ziggy sbatch ' + job + 'slurm', shell=True)
        status = outp.decode()[:-1]
        print(status)
        JobIDs.append(status.split('job ')[1])

    print(str(len(GausFiles)) + ' jobs submitted to the queue on ziggy')

    outp = subprocess.check_output('ssh ziggy qstat', shell=True)
    if settings.user in outp.decode():
        print("Jobs are running on ziggy")

    time.sleep(60)
    OldQRes = CheckZiggyQueue(JobIDs, settings)
    print('Pending: ' + str(OldQRes[0]) + ', Running: ' +
          str(OldQRes[1]) + ', Not in queue: ' + str(OldQRes[2]))

    Jobs2Complete = list(GausFiles)
    n2complete = len(Jobs2Complete)

    # Check and report on the progress of calculations
    while len(Jobs2Complete) > 0:
        JustCompleted = [job for job in Jobs2Complete if
                         IsZiggyGComplete(job[:-3] + 'out', folder, settings)]
        Jobs2Complete[:] = [job for job in Jobs2Complete if
                            not IsZiggyGComplete(job[:-3] + 'out', folder, settings)]
        if n2complete != len(Jobs2Complete):
            n2complete = len(Jobs2Complete)
            print(str(n2complete) + " remaining.")

        QRes = CheckZiggyQueue(JobIDs, settings)
        if QRes != OldQRes:
            OldQRes = QRes
            print('Pending: ' + str(OldQRes[0]) + ', Running: ' +
                  str(OldQRes[1]) + ', Not in queue: ' + str(OldQRes[2]))

        if QRes[2] == len(JobIDs):
            # check each gaussian file to ascertain the status of individual gaus jobs
            print('No jobs left in Ziggy queue')
            break

        time.sleep(120)

    # When done, copy the results back
    print("\nCopying the output files back to localhost...")
    print('ssh ziggy scp /home/' + settings.user + '/' + folder + '/*.out ' +
          socket.getfqdn() + ':' + os.getcwd())
    outp = subprocess.check_output('ssh ziggy scp /home/' + settings.user +
                                   '/' + folder + '/*.out ' + socket.getfqdn()
                                   + ':' + os.getcwd(), shell=True)


def WriteSubScript(GausJob, queue, ZiggyJobFolder, settings):

    if not (os.path.exists(GausJob)):
        print("The input file " + GausJob + " does not exist. Exiting...")
        return

    # Create the submission script
    QSub = open(GausJob[:-4] + "slurm", 'w')

    # Choose the queue
    QSub.write('#!/bin/bash\n\n')
    QSub.write('#SBATCH -p ' + settings.queue + '\n')
    if settings.nProc > 1:
        QSub.write('#SBATCH --nodes=1\n#SBATCH --cpus-per-task=' +
                   str(settings.nProc) + '\n')
    else:
        QSub.write('#SBATCH --nodes=1\n#SBATCH --cpus-per-task=1\n')
    QSub.write('#SBATCH --time=' +
               format(settings.TimeLimit, "02") + ':00:00\n\n')

    # define input files and output files
    QSub.write('file=' + GausJob[:-4] + '\n\n')
    QSub.write('inpfile=${file}.com\noutfile=${file}.out\n')

    # define cwd and scratch folder and ask the machine
    # to make it before running the job
    QSub.write('HERE=/home/' + settings.user + '/' + ZiggyJobFolder + '\n')
    QSub.write('SCRATCH=/scratch/' + settings.user + '/' +
               GausJob[:-4] + '\n')
    QSub.write('mkdir ${SCRATCH}\n')

    # Setup GAUSSIAN environment variables
    QSub.write('set OMP_NUM_THREADS=$SLURM_CPUS_PER_TASK\n')

    QSub.write('export GAUSS_EXEDIR=/usr/local/shared/gaussian/em64t/09-D01/g09\n')
    QSub.write('export g09root=/usr/local/shared/gaussian/em64t/09-D01\n')
    QSub.write('export PATH=/usr/local/shared/gaussian/em64t/09-D01/g09:$PATH\n')
    QSub.write('export GAUSS_SCRDIR=$SCRATCH\n')
    QSub.write('exe=$GAUSS_EXEDIR/g09\n')
    # copy the input file to scratch
    QSub.write('cp ${HERE}/${inpfile}  $SCRATCH\ncd $SCRATCH\n')

    # write useful info to the job output file (not the gaussian)
    QSub.write('echo "Starting job $SLURM_JOBID"\necho\n')
    QSub.write('echo "SLURM assigned me this node:"\nsrun hostname\necho\n')

    QSub.write('ln -s $HERE/$outfile $SCRATCH/$outfile\n')
    QSub.write('srun $exe > $outfile < $inpfile\n')

    # Cleanup
    QSub.write('rm -rf ${SCRATCH}/\n')

    QSub.close()


def IsZiggyGComplete(f, folder, settings):

    path = '/home/' + settings.user + '/' + folder + '/'
    try:
        outp1 = subprocess.check_output('ssh ziggy ls ' + path, shell=True)
    except subprocess.CalledProcessError as e:
        print("ssh ziggy ls failed: " + str(e.output))
        return False
    if f in outp1.decode():
        try:
            outp2 = subprocess.check_output('ssh ziggy cat ' + path + f,
                                            shell=True)
        except subprocess.CalledProcessError as e:
            print("ssh ziggy cat failed: " + str(e.output).decode())
            return False
        if ("Normal termination" in outp2[-90:].decode()) or \
                (('termination' in outp2[-270:].decode()) and ('l9999.exe' in outp2[-270:].decode())):
            return True
    return False


def CheckZiggyQueue(JobIDs, settings):

    outp = subprocess.Popen(['ssh', 'ziggy', 'qstat', '-u ' + settings.user],
                            stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()[0]
    outp = outp.decode().split('\n')

    QStart = 0
    for i, line in enumerate(outp):
        if '------' in line:
            QStart = i+1
            break
    QueueReport = outp[QStart:-1]

    JobStats = []

    for job in JobIDs:
        status = ''
        for i, line in enumerate(QueueReport):
            if job in line:
                status = list(filter(None, line.split(' ')))[9]
        JobStats.append(status)

    Pending = JobStats.count('Q')
    Running = JobStats.count('R')
    NotInQueue = JobStats.count('')

    return Pending, Running, NotInQueue



import argparse

def main():
    parser = argparse.ArgumentParser(description="Example CLI with an integer argument.")
    parser.add_argument('number', type=int, help='An integer argument')

    args = parser.parse_args()
    result = perform_operation(args.number)
    print(f"Result of the operation: {result}")

def perform_operation(number):
    # You can perform your computation or operation here
    return number * 2

if __name__ == "__main__":
    main()



import logging
import os
import shutil
import sys
import time
import pathlib
import subprocess
import re

from dp5.dft import BaseDFTMethod

logger = logging.getLogger(__name__)


class DFTMethod(BaseDFTMethod):
    def __init__(self, settings):
        super().__init__(settings)

        self.tag = "nw"
        self.input_format = ".nw"
        self.output_format = ".nwo"

        self.executable = settings["executable"]["nwchem"]

        if shutil.which(self.executable) is None:
            logger.warning(f"NWChem executable not found at {self.executable}")
            self.executable = ""

    functional_dict = {
        "b3lyp": "b3lyp",
        "m062x": "m06-2x",
        "mpw1pw91": "mpw91 perdew91",
    }

    def __repr__(self) -> str:
        return "NWChem"

    def write_file(self, filename, coordinates, atoms, charge, calc_type):
        with open(f"{filename}.nw", "w") as f:
            f.write("memory stack 1500 mb heap 1500 mb global 3000 mb\n")
            # scratch folder left for hpc
            f.write(f'echo\n\nstart molecule\n\ntitle "{filename}"\n')
            f.write("echo\n\nstart\n\n")
            if self.charge:
                f.write(f"charge {self.charge}\n\n")
            else:
                f.write(f"charge {charge}\n\n")
            f.write("geometry units angstroms print xyz autosym\n")
            for atom, (x, y, z) in zip(atoms, coordinates):
                f.write(f"  {atom} {x} {y} {z}\n")

            gaus_to_nwchem_basis = {"6-31g(d,p)": "6-31g**", "6-311g(d)": "6-311g*"}

            if calc_type == "nmr":
                basis = self.settings["n_basis_set"]
            elif calc_type == "opt":
                basis = self.settings["o_basis_set"]
            elif calc_type == "e":
                basis = self.settings["e_basis_set"]

            basis = gaus_to_nwchem_basis.get(basis.lower(), basis)

            f.write(f"end\n\nbasis\n  * library {basis}\nend\n\n")

            gaus_to_nwchem_solvents = {
                "chloroform": "chcl3",
                "dimethylsulfoxide": "dmso",
            }

            solvent = gaus_to_nwchem_solvents.get(
                self.settings["solvent"], self.settings["solvent"]
            )
            if solvent:
                f.write(f"cosmo\n  do_cosmo_smd true\n  solvent {solvent}\nend\n\n")

            if calc_type == "nmr":
                f.write(self.nmr_options())
            elif calc_type == "opt":
                f.write(self.opt_options())
            elif calc_type == "e":
                f.write(self.e_options())

    def nmr_options(self):
        functional = self.settings["n_functional"]
        functional = self.functional_dict.get(functional.lower(), functional)

        suffix = (
            f"dft\n  xc {functional}\n  mult 1\nend\n"
            f"task dft energy\n\nproperty\n  shielding\nend\ntask dft property"
        )
        return suffix

    def opt_options(self):
        functional = self.settings["o_functional"]
        functional = self.functional_dict.get(functional.lower(), functional)

        suffix = f"dft\n  xc {functional}\n  mult 1\nend\n" f"task dft optimize\n\n"
        return suffix

    def e_options(self):
        functional = self.settings["e_functional"]
        functional = self.functional_dict.get(functional.lower(), functional)

        suffix = (
            f"dft\n  xc {functional}\n  mult 1\nend\n"
            f"task dft energy\n\nproperty\n  shielding\nend\ntask dft property"
        )
        return suffix

    def prepare_command(self, file):
        return super().prepare_command(file)

    def read_file(self, file):

        atoms = []
        coordinates = []
        energy = None
        shieldings = []
        shielding_labels = []
        completed = False
        opt_converged = False

        with open(file, "r") as f:
            while not completed:
                try:
                    line = next(f)

                    if 'Geometry "geometry"' in line:
                        atoms = []
                        coordinates = []
                        for _ in range(7):
                            line = next(f)
                        while len(line) >= 2:
                            _, atom, _, x, y, z = line.split()
                            atoms.append(atom)
                            coordinates.append([float(i) for i in (x, y, z)])
                            line = next(f)

                    if "Chemical Shielding" in line:
                        while not line.strip().startswith("Task"):
                            line = next(f).strip()
                            if line.startswith("Atom:"):
                                _, at_num, at_type = line.split()
                                shielding_labels.append(f"{at_type}{at_num}")
                            if line.startswith("isotropic"):
                                *_, shielding = line.split()
                                shieldings.append(float(shielding))

                    if "Total DFT energy" in line:
                        start = line.index("Total")
                        energy = float(line[start + 19 :])

                    if "Optimization converged" in line:
                        opt_converged = True

                    if "AUTHORS" in line:
                        completed = True

                except StopIteration:
                    logger.error(
                        f"Calculations in {file} have not terminated normally."
                    )

        return (
            atoms,
            coordinates,
            energy,
            shieldings,
            shielding_labels,
            completed,
            opt_converged,
        )



import logging

from .CNN_model import get_shifts_and_labels


logger = logging.getLogger(__name__)


def get_nn_shifts(mols, batch_size=16):
    """
    Predicts shifts from rdkit Mol objects.
    Arguments:
    - list of lists of RDKit mol objects
    Returns:
    - list of lists of 13C chemical shifts for each atom in a molecule
    - list of lists of C atomic labels
    - list of lists of 1H chemical shifts for each atom in a molecule
    - list of lists of H atomic labels
    """

    C_shifts, C_labels = predict_C_shifts(mols, batch_size)

    H_shifts, H_labels = predict_H_shifts(mols, batch_size)

    # will add H_shifts and H_labels later!

    return C_shifts, C_labels, H_shifts, H_labels


def predict_C_shifts(mols, batch_size):
    return get_shifts_and_labels(
        mols,
        atomic_symbol="C",
        model_path="NMRdb-CASCADEset_Exp_mean_model_atom_features256.hdf5",
        batch_size=batch_size,
    )


def predict_H_shifts(mols, batch_size):
    return [[[]]] * len(mols), [[]] * len(mols)



import copy

import numpy as np
from scipy.optimize import linear_sum_assignment as optimise
from scipy.stats import linregress, norm

from dp5.nmr_processing.helper_functions import methyl_protons


def iterative_assignment(mol, exp_peaks, calculated_shifts, H_labels, rounded_integrals):
    """
    Arguments:
    mol: rdkit Mol object. Required to store the molecular conectivity
    exp_peaks: experimental signals as processed by NMR-AI
    calculated_shifts: simulated shifts using DFT or NN
    H_labels: atomic labels for H atoms. Must follow SD File numbering
    rounded_integrals: integrals as processed by NMR-AI

    returns:
    - assigned shifts: unused
    - assigned peaks
    - assigned labels
    - sclaed_shifts: unused
    """

    calculated_shifts = np.array(calculated_shifts)

    H_labels = np.array(H_labels)

    lnum = 0

    new_assigned_shifts = []
    old_assigned_shifts = [1]

    while old_assigned_shifts != new_assigned_shifts:

        if lnum == 0:
            # is that scaling necessary? Looks like DFT calcs
            scaled_shifts = external_scale_proton_shifts(calculated_shifts)

            scaled_mu = 0

            scaled_std = 1

        else:
            old_assigned_shifts = copy.copy(new_assigned_shifts)
            old_assigned_peaks = copy.copy(new_assigned_peaks)

            scaled_shifts, slope, intercept = internal_scale_proton_shifts(
                old_assigned_shifts, old_assigned_peaks, calculated_shifts)

            scaled_std = 1

        # assign methyl groups first

        # find methyl groups

        m_protons = methyl_protons(mol)

        m_shifts = np.array([])

        # find the average shifts of these groups

        for m_group in m_protons:

            s = 0

            for proton in m_group:

                w = np.where(H_labels == proton)

                s += scaled_shifts[w]/3

            m_shifts = np.hstack((m_shifts, s))

        # find peaks these can be assigned too

        methyl_peaks = []

        rounded_integrals = np.array(rounded_integrals)

        w = (rounded_integrals - (rounded_integrals % 3)) // 3

        for ind, peak in enumerate(sorted(list(set(exp_peaks)))[::-1]):
            methyl_peaks += [peak] * w[ind]

        # create difference matrix

        diff_matrix = np.zeros((len(m_shifts), len(methyl_peaks)))

        for ind1, i in enumerate(m_shifts):
            for ind2, j in enumerate(methyl_peaks):
                diff_matrix[ind1, ind2] = j-i

        prob_matrix = proton_probabilities(diff_matrix, scaled_mu, scaled_std)

        prob_matrix = prob_matrix**2

        prob_matrix = 1 - prob_matrix

        vertical_ind, horizontal_ind = optimise(prob_matrix)

        # unpack this assignment

        opt_labelsm = []

        opt_shiftsm = []

        opt_peaksm = []

        for j in vertical_ind:

            opt_labelsm.extend(m_protons[j])

        for i in horizontal_ind:

            opt_peaksm += 3*[methyl_peaks[i]]

        for label in opt_labelsm:

            w = np.where(H_labels == label)

            opt_shiftsm.append(calculated_shifts[w][0])

        # remove shifts/peaks/labels for the list to assign

        calculated_shiftsp = copy.copy(calculated_shifts)

        exp_peaksp = copy.copy(exp_peaks)

        scaled_shiftsp = copy.copy(scaled_shifts)

        H_labelsp = copy.copy(H_labels)

        # peaks

        for p in opt_peaksm:

            w = np.where(exp_peaksp == p)[0][0]

            exp_peaksp = np.delete(exp_peaksp, w)

        # shifts

        for s in opt_shiftsm:

            w = np.where(calculated_shiftsp == s)[0][0]

            calculated_shiftsp = np.delete(calculated_shiftsp, w)
            scaled_shiftsp = np.delete(scaled_shiftsp, w)

        # labels

        for l in opt_labelsm:

            w = np.where(H_labelsp == l)[0][0]

            H_labelsp = np.delete(H_labelsp, w)

        # assigned everything else

        diff_matrix = np.zeros((len(calculated_shiftsp), len(exp_peaksp)))

        for ind1, i in enumerate(scaled_shiftsp):
            for ind2, j in enumerate(exp_peaksp):
                diff_matrix[ind1, ind2] = j-i

        prob_matrix = proton_probabilities(diff_matrix, scaled_mu, scaled_std)

        b = abs(diff_matrix) >= 1

        # find any rows that are all zeros

        b = np.where(np.sum(prob_matrix, 1) == 0)

        prob_matrix[b] = - np.inf

        prob_matrix = np.delete(prob_matrix, b, 0)

        unassignable_shifts = calculated_shiftsp[b]

        ccalculated_shiftsp = np.delete(calculated_shiftsp, b)

        ##############################

        prob_matrix = prob_matrix**2

        prob_matrix = 1 - prob_matrix

        vertical_ind, horizontal_ind = optimise(prob_matrix)

        opt_peaksp = exp_peaksp[horizontal_ind]

        opt_shiftsp = ccalculated_shiftsp[vertical_ind]

        opt_labelsp = H_labelsp[vertical_ind]

        opt_shifts, opt_peaks, opt_labels = removecrossassignments(
            opt_peaksp, opt_shiftsp, opt_labelsp)

        # combine these assignments

        opt_peaks = np.hstack((opt_peaksm, opt_peaksp))

        opt_shifts = np.hstack((opt_shiftsm, opt_shiftsp))

        opt_labels = np.hstack((opt_labelsm, opt_labelsp))

        # check for any shifts that have not been assigned

        copyshifts = list(copy.copy(calculated_shifts))
        copylabels = list(copy.copy(H_labels))

        for shift, label in zip(opt_shifts, opt_labels):

            copyshifts.remove(shift)
            copylabels.remove(label)

        # assign these to the closest peaks - regardless of integrals

        for shift, label in zip(copyshifts, copylabels):

            mindiff = np.array(exp_peaks - shift).argmin()

            opt_peaks = np.append(opt_peaks, exp_peaks[mindiff])

            opt_labels = np.append(opt_labels, label)

            opt_shifts = np.append(opt_shifts, shift)

        # sort output wrt original H labels

        indv = []

        for label in opt_labels:

            wh = np.where(H_labels == label)

            indv.append(wh[0][0])

        ind = np.argsort(opt_shifts)[::-1]

        assigned_shifts = opt_shifts[indv]

        assigned_peaks = opt_peaks[indv]

        assigned_labels = opt_labels[indv]

        ind = np.argsort(assigned_shifts)

        assigned_shifts = assigned_shifts[ind].tolist()
        assigned_peaks = assigned_peaks[ind].tolist()
        assigned_labels = assigned_labels[ind].tolist()

        lnum += 1

        new_assigned_shifts = copy.copy(assigned_shifts)
        new_assigned_peaks = copy.copy(assigned_peaks)

    return assigned_shifts, assigned_peaks, assigned_labels, scaled_shifts


def proton_probabilities(diff_matrix, scaled_mu, scaled_std):

    prob_matrix = norm.pdf(diff_matrix, scaled_mu, scaled_std) / \
        norm.pdf(scaled_mu, scaled_mu, scaled_std)

    return prob_matrix


def removecrossassignments(exp, calc, labels):

    # sort these in decending order

    s = np.argsort(calc)[::-1]

    calc = calc[s]

    exp = exp[s]

    labels = labels[s]

    # generate difference matrix
    switch = 0

    expcopy = np.array(exp)

    while switch == 0:

        swapm = np.zeros([len(calc), len(calc)])

        for i, Hi in enumerate(expcopy):
            for j, Hj in enumerate(expcopy):

                if i > j:

                    swapm[i, j] = 0
                else:
                    swapm[i, j] = round(Hi - Hj, 1)

        w = np.argwhere(swapm < 0)

        if len(w > 0):
            expcopy[w[0]] = expcopy[w[0][::-1]]

        else:
            switch = 1

    return calc, expcopy, labels


def external_scale_proton_shifts(calculated_shifts):
    """
    - Calculated_shifts(numpy array)
    may not be true, depends on particular basis set"""
    scaled = 0.9770793502768845 * calculated_shifts - 0.019505417520415236

    return scaled


def internal_scale_proton_shifts(assigned_shifts, assigned_peaks, calculated_shifts):

    slope, intercept, r_value, p_value, std_err = linregress(
        assigned_shifts, assigned_peaks)

    scaled_shifts = calculated_shifts * slope + intercept

    return scaled_shifts, slope, intercept



import itertools
import copy

from matplotlib import pyplot as plt
import numpy as np
from lmfit import Minimizer, Parameters, report_fit
import nmrglue as ng
from scipy.stats import gmean, norm
from scipy.optimize import linear_sum_assignment as optimise
from scipy.interpolate import InterpolatedUnivariateSpline
import pathos.multiprocessing as mp
from rdkit import Chem

def process_proton(NMR_file, settings, datatype):

    total_spectral_ydata, spectral_xdata_ppm, corr_distance, uc, noise_std, peak_regions = spectral_processing(NMR_file,
                                                                                                               datatype)

    gradient_peaks, gradient_regions, gradient_groups, std = gradient_peak_picking(total_spectral_ydata, corr_distance,
                                                                                   uc, noise_std, peak_regions)

    import time

    start = time.time()

    picked_peaks, grouped_peaks, peak_regions, sim_y, total_params = multiproc_BIC_minimisation(gradient_regions,
                                                                                                gradient_groups,
                                                                                                total_spectral_ydata,
                                                                                                corr_distance,
                                                                                                uc, noise_std)

    end = time.time()

    print("minimisation time = " + str((end - start) / 60) + " mins")

    peak_regions, picked_peaks, grouped_peaks, spectral_xdata_ppm, solvent_region_ind = editsolvent_removal2(
        settings.Solvent, total_spectral_ydata, spectral_xdata_ppm, picked_peaks, peak_regions, grouped_peaks,
        total_params,
        uc)

    sim_regions, full_sim_data = simulate_regions(total_params, peak_regions, grouped_peaks, total_spectral_ydata,
                                                  spectral_xdata_ppm)

    peak_regions, grouped_peaks, sim_regions, integral_sum, cummulative_vectors, integrals, number_of_protons_structure, optimum_proton_number, total_integral = find_integrals(

        settings.InputFiles[0], peak_regions, grouped_peaks, sim_regions, picked_peaks, total_params,
        total_spectral_ydata, solvent_region_ind)

    # find region centres

    centres = weighted_region_centres(peak_regions, total_spectral_ydata)

    ################

    exp_peaks = []

    integrals = np.array([int(i) for i in integrals])

    for ind, peak in enumerate(centres):
        exp_peaks += [peak] * integrals[ind]

    integrals = integrals[integrals > 0.5]

    exp_peaks = spectral_xdata_ppm[exp_peaks]

    exp_peaks = np.array([ round(i,4) for i in exp_peaks])

    return exp_peaks, spectral_xdata_ppm, total_spectral_ydata, integrals, peak_regions, centres, cummulative_vectors, integral_sum, picked_peaks, total_params, sim_regions


def guess_udic(dic, data):
    """
    Guess parameters of universal dictionary from dic, data pair.
    Parameters
    ----------
    dic : dict
        Dictionary of JCAMP-DX parameters.
    data : ndarray
        Array of NMR data.
    Returns
    -------
    udic : dict
        Universal dictionary of spectral parameters.
    """

    # create an empty universal dictionary
    udic = ng.fileiobase.create_blank_udic(1)

    # update default values (currently only 1D possible)
    # "label"
    try:
        label_value = dic[".OBSERVENUCLEUS"][0].replace("^", "")
        udic[0]["label"] = label_value
    except KeyError:
        # sometimes INSTRUMENTAL PARAMETERS is used:
        try:
            label_value = dic["INSTRUMENTALPARAMETERS"][0].replace("^", "")
            udic[0]["label"] = label_value
        except KeyError:
            pass

    # "obs"
    obs_freq = None
    try:
        obs_freq = float(dic[".OBSERVEFREQUENCY"][0])
        udic[0]["obs"] = obs_freq

    except KeyError:
        pass

    # "size"
    if isinstance(data, list):
        data = data[0]  # if list [R,I]
    if data is not None:
        udic[0]["size"] = len(data)

    # "sw"
    # get firstx, lastx and unit
    firstx, lastx, isppm = ng.jcampdx._find_firstx_lastx(dic)

    # ppm data: convert to Hz
    if isppm:
        if obs_freq:
            firstx = firstx * obs_freq
            lastx = lastx * obs_freq
        else:
            firstx, lastx = (None, None)

    if firstx is not None and lastx is not None:
        udic[0]["sw"] = abs(lastx - firstx)

        # keys not found in standard&required JCAMP-DX keys and thus left default:
        # car, complex, encoding

        udic[0]['car'] = firstx - abs(lastx - firstx) / 2

    return udic


def spectral_processing(file, datatype):
    print('Processing Proton Spectrum')

    if datatype == 'jcamp':

        dic, total_spectral_ydata = ng.jcampdx.read(file)  # read file

        total_spectral_ydata = total_spectral_ydata[0] + 1j * total_spectral_ydata[1]

        total_spectral_ydata = ng.proc_base.ifft_positive(total_spectral_ydata)

    else:

        dic, total_spectral_ydata = ng.bruker.read(file)  # read file

        total_spectral_ydata = ng.bruker.remove_digital_filter(dic, total_spectral_ydata)  # remove the digital filter

    total_spectral_ydata = ng.proc_base.zf_double(total_spectral_ydata, 4)

    total_spectral_ydata = ng.proc_base.fft_positive(total_spectral_ydata)  # Fourier transform

    corr_distance = estimate_autocorrelation(total_spectral_ydata)

    # normalise the data

    m = max(np.max(abs(np.real(total_spectral_ydata))), np.max(abs(np.imag(total_spectral_ydata))))

    total_spectral_ydata = np.real(total_spectral_ydata / m) + 1j * np.imag(total_spectral_ydata / m)

    if datatype == 'jcamp':

        udic = guess_udic(dic, total_spectral_ydata)

    else:

        udic = ng.bruker.guess_udic(dic, total_spectral_ydata)  # sorting units

    uc = ng.fileiobase.uc_from_udic(udic)  # unit conversion element

    spectral_xdata_ppm = uc.ppm_scale()  # ppmscale creation

    # baseline and phasing

    tydata = ACMEWLRhybrid(total_spectral_ydata, corr_distance)

    # find final noise distribution
    classification, sigma = baseline_find_signal(tydata, corr_distance, True, 1)

    # fall back phasing if fit doesnt converge

    # calculate negative area

    # draw regions

    peak_regions = []

    c1 = np.roll(classification, 1)

    diff = classification - c1

    s_start = np.where(diff == 1)[0]

    s_end = np.where(diff == -1)[0] - 1

    for r in range(len(s_start)):
        peak_regions.append(np.arange(s_start[r], s_end[r]))

    tydata = tydata / np.max(abs(tydata))

    return tydata, spectral_xdata_ppm, corr_distance, uc, sigma, peak_regions


def estimate_autocorrelation(total_spectral_ydata):
    # note this region may have a baseline distortion

    y = np.real(total_spectral_ydata[0:10000])

    params = Parameters()

    # define a basleine polnomial

    order = 6

    for p in range(order + 1):
        params.add('p' + str(p), value=0)

    def poly(params, order, y):

        bl = np.zeros(len(y))
        x = np.arange(len(y))

        for p in range(order + 1):
            bl += params['p' + str(p)] * x ** (p)

        return bl

    def res(params, order, y):

        bl = poly(params, order, y)

        r = abs(y - bl)

        return r

    out = Minimizer(res, params,
                    fcn_args=(order, y))

    results = out.minimize()

    bl = poly(results.params, order, y)

    y = y - bl

    t0 = np.sum(y * y)

    c = 1

    tc = 1

    t = []

    while tc > 0.36:
        tc = np.sum(np.roll(y, c) * y) / t0

        t.append(tc)

        c += 1

    return c


def acme(y, corr_distance):
    params = Parameters()

    phase_order = 3

    for p in range(phase_order + 1):
        params.add('p' + str(p), value=0, min=-np.pi, max=np.pi)

    def acmescore(params, im, real, phase_order):

        """
        Phase correction using ACME algorithm by Chen Li et al.
        Journal of Magnetic Resonance 158 (2002) 164-168

        Parameters
        ----------
        pd : tuple
            Current p0 and p1 values
        data : ndarray
            Array of NMR data.

        Returns
        -------
        score : float
            Value of the objective function (phase score)

        """

        data = ps(params, im, real, phase_order)

        ##########

        # calculate entropy of non corrected data, calculate penalty for baseline corrected data
        #  - keep as vector to use the default lmfit method

        # Calculation of first derivatives of signal regions

        ds1 = np.abs((data[1:] - data[:-1]))

        p1 = ds1 / np.sum(ds1)

        # Calculation of entropy
        p1[p1 == 0] = 1

        h1 = -p1 * np.log(p1)
        # h1s = np.sum(h1)

        # Calculation of penalty
        pfun = 0.0
        as_ = data - np.abs(data)
        # as_ = databl - np.abs(databl)
        sumas = np.sum(as_)

        if sumas < 0:
            # pfun = pfun + np.sum((as_ / 2) ** 2)
            pfun = (as_[1:] / 2) ** 2

        p = 1000 * pfun

        return h1 + p

    out = Minimizer(acmescore, params,
                    fcn_args=(np.imag(y), np.real(y), phase_order))

    results = out.minimize()

    p = results.params

    p.pretty_print()

    y = ps(p, np.imag(y), np.real(y), phase_order)

    classification, sigma = baseline_find_signal(y, corr_distance, True, 1)
    r = gen_baseline(np.real(y), classification, corr_distance)
    y -= r

    return y


def ACMEWLRhybrid(y, corr_distance):

    def residual_function(params, im, real):

        # phase the region

        data = ps(params, im, real, 0)

        # make new baseline for this region

        r = np.linspace(data[0], data[-1], len(real))

        # find negative area

        data -= r

        ds1 = np.abs((data[1:] - data[:-1]))

        p1 = ds1 / np.sum(ds1)

        # Calculation of entropy
        p1[p1 == 0] = 1

        h1 = -p1 * np.log(p1)
        h1s = np.sum(h1)

        # Calculation of penalty
        pfun = 0.0

        as_ = data - np.abs(data)

        sumas = np.sum(as_)

        if sumas < 0:
            pfun = (as_[1:] / 2) ** 2

        p = np.sum(pfun)

        return h1s + 1000 * p

    # find regions

    classification, sigma = baseline_find_signal(y, corr_distance, True, 1)

    c1 = np.roll(classification, 1)

    diff = classification - c1

    s_start = np.where(diff == 1)[0]

    s_end = np.where(diff == -1)[0] - 1

    peak_regions = []

    for r in range(len(s_start)):
        peak_regions.append(np.arange(s_start[r], s_end[r]))

    # for region in peak_regions:
    #    plt.plot(region,y[region],color = 'C1')

    # phase each region independently

    phase_angles = []

    weights = []

    centres = []

    for region in peak_regions:
        params = Parameters()

        params.add('p0', value=0, min=-np.pi, max=np.pi)

        out = Minimizer(residual_function, params,
                        fcn_args=(np.imag(y[region]), np.real(y[region])))

        results = out.minimize('brute')

        p = results.params

        phase_angles.append(p['p0'] * 1)

        # find weight

        data = ps(p, np.imag(y[region]), np.real(y[region]), 0)

        # make new baseline for this region

        r = np.linspace(data[0], data[-1], len(data))

        # find negative area

        res = data - r

        weights.append(abs(np.sum(res[res > 0] / np.sum(y[y > 0]))))

        centres.append(np.median(region) / len(y))

    sw = sum(weights)

    weights = [w / sw for w in weights]

    # do weighted linear regression on the regions

    # do outlier analysis

    switch = 0

    centres = np.array(centres)

    weights = np.array(weights)

    sweights = np.argsort(weights)[::-1]

    phase_angles = np.array(phase_angles)

    ind1 = 0

    while switch == 0:

        intercept, gradient = np.polynomial.polynomial.polyfit(centres, phase_angles, deg=1, w=weights)

        predicted_angles = gradient * centres + intercept

        weighted_res = np.abs(predicted_angles - phase_angles) * weights

        # find where largest weighted residual is

        max_res = sweights[ind1]

        s = 0

        if phase_angles[max_res] > 0:

            s = -1

            phase_angles[max_res] -= 2 * np.pi

        else:

            s = +1

            phase_angles[max_res] += 2 * np.pi

        intercept1, gradient1 = np.polynomial.polynomial.polyfit(centres, phase_angles, deg=1, w=weights)

        new_predicted_angles = gradient1 * centres + intercept1

        new_weighted_res = np.abs(new_predicted_angles - phase_angles) * weights

        if np.sum(new_weighted_res)  > np.sum(weighted_res):

            switch = 1

            phase_angles[max_res] += -2*np.pi*s

        ind1 +=1

    # phase the data

    p_final = Parameters()

    p_final.add('p0', value=intercept)
    p_final.add('p1', value=gradient)

    # p_final.pretty_print()

    y = ps(p_final, np.imag(y), np.real(y), 1)


    classification, sigma = baseline_find_signal(y, corr_distance, True, 1)
    r = gen_baseline(np.real(y), classification, corr_distance)
    y -= r

    return np.real(y)


def ps(param, im, real, phase_order):
    x = np.linspace(0, 1, len(real))

    angle = np.zeros(len(x))

    for p in range(phase_order + 1):
        angle += param['p' + str(p)] * x ** (p)

    # phase the data

    R = real * np.cos(angle) - im * np.sin(angle)

    return R


def baseline_find_signal(y_data, cdist, dev, t):
    wd = int(cdist) * 10

    sd_all = _get_sd(y_data, wd)

    snvectort = np.zeros(len(y_data))

    sv = []

    for i in range(0, 4 * cdist):
        x = np.arange(i + wd, len(y_data) - wd, 4 * cdist)

        sample = y_data[x]

        sd_set = _get_sd(sample, wd)

        s = _find_noise_sd(sd_set, 0.999)

        sv.append(s)

    sigma = np.mean(sv)

    b = np.linspace(-0.001, 0.001, 1000)

    if dev == True:

        w = np.where(sd_all > t * sigma)[0]

    else:

        w = np.where(y_data > t * sigma)[0]

    snvectort[w] = 1

    sn_vector = np.zeros(len(y_data))

    w = cdist

    for i in np.arange(len(sn_vector)):
        if snvectort[i] == 1:
            sn_vector[np.maximum(0, i - w):np.minimum(i + w, len(sn_vector))] = 1

    return sn_vector, sigma


def gen_baseline(y_data, sn_vector, corr_distance):
    points = np.arange(len(y_data))

    spl = InterpolatedUnivariateSpline(points[sn_vector == 0], y_data[sn_vector == 0], k=1)

    r = spl(points)

    # is corr distance odd or even

    if corr_distance % 2 == 0:
        kernel = np.ones((corr_distance + 1) * 10) / ((corr_distance + 1) * 10)
    else:
        kernel = np.ones((corr_distance) * 10) / ((corr_distance) * 10)

    r = np.convolve(r, kernel, mode='same')

    return r


def _rolling_window(a, window):
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


def _get_sd(data, k):
    return np.std(_rolling_window(data, k), -1)


def _find_noise_sd(sd_set, ratio):
    '''Calculate the median m1 from SDset. exclude the elements greater
    than 2m1from SDset and recalculate the median m2. Repeat until
    m2/m1 converge(sd_set)'''
    m1 = np.median(sd_set)
    S = sd_set <= 2.0 * m1
    tmp = S * sd_set
    sd_set = tmp[tmp != 0]
    m2 = np.median(sd_set)
    while m2 / m1 < ratio:
        m1 = np.median(sd_set)
        S = sd_set <= 2.0 * m1
        tmp = S * sd_set
        sd_set = tmp[tmp != 0]

    return m2


########################################################################################################################
# peak picking and minimisation
########################################################################################################################


def p7(x, mu, std, v, A):
    x1 = (mu - x) / (std / 2)

    y = (1 - v) * (1 / (1 + (x1) ** 2)) + (v) * ((1 + ((x1) ** 2) / 2) / (1 + (x1) ** 2 + (x1) ** 4))

    y *= A

    return y


def p7residual(params, x, picked_points, y_data, region, differential):
    y = np.zeros(len(x))

    for peak in picked_points:
        y += p7(x, params['mu' + str(peak)], params['std' + str(peak)], params['vregion' + str(region)],
                params['A' + str(peak)])

    if differential == True:

        res = abs(y - y_data)

        av = np.average(res)

        av2 = np.average(res ** 2)

        difference = (res ** 2 - av2) - (res - av) ** 2

    else:
        difference = (y - y_data) ** 2

    return difference


def p7residualsolvent(params, x, picked_points, y_data, region, differential):
    y = np.zeros(len(x))

    for peak in picked_points:
        y += p7(x, params['mu' + str(peak)], params['std' + str(peak)], params['vregion' + str(region)],
                params['A' + str(peak)] * params['global_amp'])

    if differential == True:

        dy = np.gradient(y)
        ddy = np.gradient(y)

        dy_ydata = np.gradient(y_data)
        ddy_ydata = np.gradient(dy_ydata)

        difference = (y - y_data) ** 2 + (dy_ydata - dy) ** 2 + (ddy_ydata - ddy) ** 2

    else:
        difference = abs(y - y_data)

    return difference


def p7simsolvent(params, x, picked_points, region):
    y = np.zeros(len(x))

    for peak in picked_points:
        y += p7(x, params['mu' + str(peak)], params['std' + str(peak)], params['vregion' + str(region)],
                params['A' + str(peak)] * params['global_amp'])

    return y


def p7sim(params, x, picked_points, region):
    y = np.zeros(len(x))

    for peak in picked_points:
        y += p7(x, params['mu' + str(peak)], params['std' + str(peak)], params['vregion' + str(region)],
                params['A' + str(peak)])

    return y


def p7plot(params, region, group, ind, xppm):
    region_j = np.zeros(len(region))

    for peak in group:
        j = p7(region, params['mu' + str(peak)], params['std' + str(peak)], params['vregion' + str(ind)],
               params['A' + str(peak)])

        region_j += j

    return region_j


############

###########

###########


def gradient_peak_picking(y_data, corr_distance, uc, std, binary_map_regions):
    final_peaks = []

    # estimate std of second derivative data

    ddy = np.diff(y_data, 2)

    ddy = ddy / np.max(ddy)

    # find peaks

    classification, sigma = baseline_find_signal(-1 * ddy, corr_distance, False, 2)

    ddy1 = np.roll(ddy, 1)

    ddyn1 = np.roll(ddy, -1)

    p = np.where((ddy < ddy1) & (ddy < ddyn1))[0]

    peaks = p[classification[p] == 1]

    peaks1 = np.roll(peaks, 1)

    distance = np.min(abs(peaks1 - peaks))

    # must make sure the convolution kernel is odd in length to prevent the movement of the peaks

    peaks = np.sort(peaks)

    peakscopy = copy.copy(peaks)
    ddycopy = copy.copy(ddy[peaks] / np.max(ddy))

    while distance < corr_distance:
        # roll the peaks one forward
        peakscopy1 = np.roll(peakscopy, 1)

        # find distances between peaks
        diff = np.abs(peakscopy - peakscopy1)

        # find where in the array the smallest distance is
        mindist = np.argmin(diff)

        # what is this distance
        distance = diff[mindist]

        # compare the values of the second derivative at the closest two peaks

        compare = np.argmax(ddycopy[[mindist, mindist - 1]])

        peakscopy = np.delete(peakscopy, mindist - compare)
        ddycopy = np.delete(ddycopy, mindist - compare)

    # remove any peaks that fall into the noise

    n = y_data[peakscopy]

    w = n > 5 * std

    peakscopy = peakscopy[w]

    final_peaks = sorted(list(peakscopy))

    # draw new regions symmetrically around the newly found peaks

    dist_hz = uc(0, "Hz") - uc(9, "Hz")

    peak_regions = []

    for peak in final_peaks:
        l = np.arange(peak + 1, min(peak + dist_hz + 1, len(y_data))).tolist()

        m = np.arange(max(peak - dist_hz, 0), peak).tolist()

        region = m + [peak] + l

        peak_regions.append(region)

    final_regions = [peak_regions[0]]
    final_peaks_seperated = [[final_peaks[0]]]

    for region in range(1, len(peak_regions)):

        if peak_regions[region][0] <= final_regions[-1][-1]:

            final_regions[-1] += peak_regions[region]

            final_peaks_seperated[-1].append(final_peaks[region])

        else:

            final_regions += [peak_regions[region]]

            final_peaks_seperated.append([final_peaks[region]])

    final_regions = [np.arange(min(region), max(region) + 1).tolist() for region in final_regions]

    return final_peaks, final_regions, final_peaks_seperated, std


def multiproc_BIC_minimisation(peak_regions, grouped_peaks, total_spectral_ydata, corr_distance, uc, std):
    maxproc = 5

    pool = mp.Pool(maxproc)

    new_grouped_peaks = [[] for i in range(len(peak_regions))]
    new_grouped_params = [[] for i in range(len(peak_regions))]
    new_sim_y = [[] for i in range(len(peak_regions))]

    def BIC_minimisation_region_full(ind1, uc, peak_regions, grouped_peaks, total_spectral_ydata, corr_distance, std):

        ################################################################################################################
        # initialise process
        ################################################################################################################

        # print("minimising region " + str(ind1) + " of " + str(len(peak_regions)))

        BIC_param = 15

        region = np.array(peak_regions[ind1])

        region_y = total_spectral_ydata[region]

        fit_y = np.zeros(len(region_y))

        copy_peaks = np.array(grouped_peaks[ind1])

        params = Parameters()

        fitted_peaks = []

        ttotal = 0

        ################################################################################################################
        # build initial model
        ################################################################################################################

        # params.add('vregion' + str(ind1), value=2.5, max=5, min=1)

        params.add('vregion' + str(ind1), value=0.5, max=1, min=0)

        distance = uc(0, "hz") - uc(5, "hz")

        std_upper = uc(0, "hz") - uc(1, "hz")
        av_std = uc(0, "hz") - uc(0.2, "hz")
        std_lower = uc(0, "hz") - uc(0.1, "hz")

        # build model

        while (len(copy_peaks) > 0):
            # pick peak that is furthest from fitted data:

            diff_array = region_y - fit_y

            ind2 = np.argmax(diff_array[copy_peaks - region[0]])

            maxpeak = copy_peaks[ind2]

            copy_peaks = np.delete(copy_peaks, ind2)

            # only allow params < distance away vary at a time

            # add new params

            fitted_peaks.append(maxpeak)

            fitted_peaks = sorted(fitted_peaks)

            params.add('A' + str(maxpeak), value=total_spectral_ydata[maxpeak], min=0, max=1, vary=True)

            # params.add('std' + str(maxpeak), value=av_std, vary=True, min = std_lower,
            #               max = std_upper)

            params.add('std' + str(maxpeak), value=av_std, vary=True)

            params.add('mu' + str(maxpeak), value=maxpeak, vary=True
                       , min=maxpeak - 4 * corr_distance, max=maxpeak + 4 * corr_distance)

            # adjust amplitudes and widths of the current model

        initial_y = p7sim(params, region, fitted_peaks, ind1)

        inty = np.sum(region_y[region_y > 0])

        intmodel = np.sum(initial_y)

        # check the region can be optimised this way

        # find peak with max amplitude

        maxamp = 0

        for peak in fitted_peaks:
            amp = params['A' + str(peak)]
            if amp > maxamp:
                maxamp = copy.copy(amp)

        maxintegral = maxamp * len(region)

        if maxintegral > inty:

            # set initial conditions

            while (intmodel / inty < 0.99) or (intmodel / inty > 1.01):

                for f in fitted_peaks:
                    params['std' + str(f)].set(value=params['std' + str(f)] * inty / intmodel)

                initial_y = p7sim(params, region, fitted_peaks, ind1)

                for f in fitted_peaks:
                    params['A' + str(f)].set(
                        value=params['A' + str(f)] * region_y[int(params['mu' + str(f)]) - region[0]] / (
                            initial_y[f - region[0]]))

                initial_y = p7sim(params, region, fitted_peaks, ind1)

                intmodel = np.sum(initial_y)

        # print('built model region ' + str(ind1))

        ################################################################################################################
        # now relax all params
        ################################################################################################################

        # allow all params to vary

        params['vregion' + str(ind1)].set(vary=True)

        for peak in fitted_peaks:
            params['A' + str(peak)].set(vary=False, min=max(0, params['A' + str(peak)] - 0.01),
                                        max=min(params['A' + str(peak)] + 0.01, 1))
            params['mu' + str(peak)].set(vary=False)
            params['std' + str(peak)].set(vary=False, min=min(std_lower, params['std' + str(peak)] - av_std),
                                          max=max(params['std' + str(peak)] + av_std, std_upper))

        out = Minimizer(p7residual, params,
                        fcn_args=(region, fitted_peaks, region_y, ind1, False))

        results = out.minimize()

        params = results.params

        # print('relaxed params region ' + str(ind1))

        ################################################################################################################
        # now remove peaks in turn
        ################################################################################################################

        trial_y = p7sim(params, region, fitted_peaks, ind1)

        trial_peaks = np.array(fitted_peaks)

        amps = []

        for peak in trial_peaks:
            amps.append(params['A' + str(peak)])

        r = trial_y - region_y

        chi2 = r ** 2

        N = len(chi2)

        BIC = N * np.log(np.sum(chi2) / N) + np.log(N) * (3 * len(fitted_peaks) + 2)

        while (len(trial_peaks) > 0):

            new_params = copy.copy(params)

            # find peak with smallest amp

            minpeak = trial_peaks[np.argmin(amps)]

            # remove this peak from the set left to try

            trial_peaks = np.delete(trial_peaks, np.argmin(amps))
            amps = np.delete(amps, np.argmin(amps))

            # remove this peak from the trial peaks list and the trial params

            new_params.__delitem__('A' + str(minpeak))
            new_params.__delitem__('mu' + str(minpeak))
            new_params.__delitem__('std' + str(minpeak))

            new_fitted_peaks = np.delete(fitted_peaks, np.where(fitted_peaks == minpeak))

            # simulate data with one fewer peak

            new_trial_y = p7sim(new_params, region, new_fitted_peaks, ind1)

            r = new_trial_y - region_y

            chi2 = np.sum(r ** 2)

            N = len(new_trial_y)

            new_BIC = N * np.log(chi2 / N) + np.log(N) * (3 * len(new_fitted_peaks) + 2)

            # if the fit is significantly better remove this peak

            if new_BIC < BIC - BIC_param:
                fitted_peaks = copy.copy(new_fitted_peaks)

                params = copy.copy(new_params)

                BIC = copy.copy(new_BIC)

        fitted_peaks = sorted(fitted_peaks)

        fit_y = p7sim(params, region, fitted_peaks, ind1)

        ################################################################################################################

        return fitted_peaks, params, fit_y

    # order regions by size to efficiently fill cores

    region_lengths = np.array([len(g) for g in grouped_peaks])

    sorted_regions = np.argsort(region_lengths)[::-1]

    res = [[] for i in peak_regions]

    # write output files

    for ind1 in range(len(peak_regions)):
        res[ind1] = pool.apply_async(BIC_minimisation_region_full,
                                     [ind1, uc, peak_regions, grouped_peaks, total_spectral_ydata, corr_distance,
                                      std])

    for ind1 in sorted_regions:
        new_grouped_peaks[ind1], new_grouped_params[ind1], new_sim_y[ind1] = res[ind1].get()

    #### unpack the parameters and split groups

    final_grouped_peaks = []
    final_peak_regions = []
    final_sim_y = []
    final_peaks = []

    total_params = Parameters()

    new_peaks = []

    dist_hz = uc(0, "Hz") - uc(20, "Hz")

    newgroupind = 0
    oldgroupind = 0

    for group in new_grouped_peaks:

        group = sorted(group)

        if len(group) > 0:
            final_grouped_peaks.append([])
            total_params.add('vregion' + str(newgroupind),
                             value=new_grouped_params[oldgroupind]['vregion' + str(oldgroupind)])

            final_peaks.extend(group)

        for ind2, peak in enumerate(group):

            # check if the group should be split

            if ind2 > 0:

                # if there is a gap of more than 20Hz split the group

                if peak > group[ind2 - 1] + dist_hz:
                    # track the numer of splits included to ensure v parameter is added to the correct group each time

                    newgroupind += 1

                    # if a split occures add a new v parameter for the new group

                    total_params.add('vregion' + str(newgroupind),
                                     value=new_grouped_params[oldgroupind]['vregion' + str(oldgroupind)])

                    # allow peaks to be added to the new group
                    final_grouped_peaks.append([])

            # finally append the peak
            final_grouped_peaks[-1].append(peak)
            total_params.add('A' + str(peak), value=new_grouped_params[oldgroupind]['A' + str(peak)])
            total_params.add('std' + str(peak), value=new_grouped_params[oldgroupind]['std' + str(peak)])
            total_params.add('mu' + str(peak), value=new_grouped_params[oldgroupind]['mu' + str(peak)], vary=False)

        if len(group) > 0:
            newgroupind += 1

        oldgroupind += 1

    # draw regions between midpoints of groups

    for ind4, group in enumerate(final_grouped_peaks):

        if ind4 == 0:
            lower_point = 0
            higher_point = int((group[-1] + final_grouped_peaks[ind4 + 1][0]) / 2)

        elif ind4 == len(final_grouped_peaks) - 1:
            lower_point = int((group[0] + final_grouped_peaks[ind4 - 1][-1]) / 2)
            higher_point = len(total_spectral_ydata)

        else:
            lower_point = int((group[0] + final_grouped_peaks[ind4 - 1][-1]) / 2)

            higher_point = int((group[-1] + final_grouped_peaks[ind4 + 1][0]) / 2)

        final_peak_regions.append(np.arange(lower_point, higher_point))

    # now simulate new regions and store region data

    for ind3, region in enumerate(final_peak_regions):
        fit_y = p7sim(total_params, region, final_grouped_peaks[ind3], ind3)

        final_sim_y.append(fit_y)

    return final_peaks, final_grouped_peaks, final_peak_regions, final_sim_y, total_params


########################################################################################################################
# integration and processing
########################################################################################################################


def lorentzian(p, w, p0, A):
    x = (p0 - p) / (w / 2)

    L = A / (1 + x ** 2)

    return L


def lorenz_curves(params, x, picked_points):
    y = np.zeros(len(x))
    for peak in picked_points:
        y += lorentzian(x, params['width' + str(peak)], params['pos' + str(peak)], params['amp' + str(peak)])
    return y


def simulate_regions(params, peak_regions, grouped_peaks, y_data, xppm):
    sim_regions = []
    sim_y = np.zeros(len(y_data))

    for ind, region in enumerate(peak_regions):
        sim_y[region] = p7plot(params, region, grouped_peaks[ind], ind, xppm)

        y = p7sim(params, region, grouped_peaks[ind], ind)

        sim_regions.append(y)

    return sim_regions, sim_y


def new_first_order_peak(start_ppm, J_vals, x_data, corr_distance, uc, spin):
    # new first order peak generator using the method presented in Hoye paper

    start = uc(str(start_ppm) + "ppm")

    start_Hz = uc.hz(start)

    J_vals = np.array(J_vals)

    peaks = np.zeros((2 * spin + 1) ** len(J_vals))

    if spin == 0.5:
        l = [1, -1]

    if spin == 1:
        l = [1, 0, -1]

    # signvector generator

    signvectors = itertools.product(l, repeat=len(J_vals))

    for ind, sv in enumerate(signvectors):
        shift = J_vals * sv

        shift = start_Hz + np.sum(shift)
        peaks[ind] = shift

    peaks = np.sort(peaks)

    peak_vector = np.array(sorted(list(set(peaks)), reverse=True))

    amp_vector = np.zeros(len(peak_vector))

    for peak in peaks:
        index = np.where(peak_vector == peak)
        amp_vector[index] += 1

    pv = []

    for index, peak in enumerate(peak_vector):
        pv.append(uc(peak, "Hz"))

    peak_vector = pv

    split_params = Parameters()

    for index, peak in enumerate(peak_vector):
        split_params.add('amp' + str(peak), value=amp_vector[index])
        split_params.add('pos' + str(peak), value=peak)
        split_params.add('width' + str(peak), value=2 * corr_distance)

    y = lorenz_curves(split_params, x_data, peak_vector)

    y = y / np.max(y)

    # where = np.where(y > 0.001)

    # y = y[where]

    return peak_vector, amp_vector, y

def editsolvent_removal2(solvent, y_data, x_data, picked_peaks, peak_regions, grouped_peaks, total_params, uc):
    picked_peaks = np.array(picked_peaks)

    # define the solvents

    if solvent == 'chloroform':

        exp_ppm = [7.26]

        Jv = [[]]

    elif solvent == 'dimethylsulfoxide':

        exp_ppm = [2.50]

        Jv = [[1.9, 1.9]]

    elif solvent == 'methanol':

        exp_ppm = [4.78, 3.31]

        Jv = [[], [1.7, 1.7]]

    elif solvent == 'benzene':

        exp_ppm = [7.16]

        Jv = [[]]

    elif solvent == 'pyridine':

        exp_ppm = [8.74, 7.58, 7.22]

        Jv = [[], [], []]

    else:
        exp_ppm = []
        Jv = [[]]

    # find picked peaks ppm values

    picked_peaks_ppm = x_data[picked_peaks]

    # make differences vector for referencing against multiple solvent peaks

    differences = []

    peaks_to_remove = []

    solvent_regions = []

    # now remove each peak in turn

    for ind1, speak_ppm in enumerate(exp_ppm):

        # if only a singlet is expected for this peak find solvent peak based on amplitude and position

        if len(Jv[ind1]) == 0:

            probs = norm.pdf(abs(picked_peaks_ppm - speak_ppm), loc=0, scale=0.1) * y_data[picked_peaks]

            # find the maximum probability

            w = np.argmax(probs)

            # append this to the list to remove

            peaks_to_remove.append(picked_peaks[w])

            # append this to the list of differences

            differences.append(speak_ppm - picked_peaks_ppm[w])

        # if the peak displays a splitting pattern then we have to be a bit more selective
        # do optimisation problem with projected peaks
        else:

            amp_res = []
            dist_res = []
            pos_res = []

            # limit the search to peaks +- 1 ppm either side

            srange = (picked_peaks_ppm > speak_ppm - 1) * (picked_peaks_ppm < speak_ppm + 1)

            for peak in picked_peaks_ppm[srange]:

                # print("picked ppm ", peak)

                fit_s_peaks, amp_vector, fit_s_y = new_first_order_peak(peak, Jv[ind1], np.arange(len(x_data)), 0.1, uc,
                                                                        1)

                diff_matrix = np.zeros((len(fit_s_peaks), len(picked_peaks)))

                for i, f in enumerate(fit_s_peaks):

                    for j, g in enumerate(picked_peaks):
                        diff_matrix[i, j] = abs(f - g)

                # minimise these distances

                vertical_ind, horizontal_ind = optimise(diff_matrix)

                closest_peaks = np.sort(picked_peaks[horizontal_ind])

                closest_amps = []

                for cpeak in closest_peaks:
                    closest_amps.append(total_params['A' + str(cpeak)])

                # find the amplitude residual between the closest peaks and the predicted pattern

                # normalise these amplitudes

                amp_vector = [i / max(amp_vector) for i in amp_vector]

                closest_amps = [i / max(closest_amps) for i in closest_amps]

                # append to the vector

                amp_res.append(sum([abs(amp_vector[i] - closest_amps[i]) for i in range(len(amp_vector))]))

                dist_res.append(np.sum(np.abs(closest_peaks - fit_s_peaks)))

                pos_res.append(norm.pdf(abs(peak - speak_ppm), loc=0, scale=0.5))

                # use the gsd data to find amplitudes of these peaks

            pos_res = [1 - i / max(pos_res) for i in pos_res]

            dist_res = [i / max(dist_res) for i in dist_res]

            amp_res = [i / max(amp_res) for i in amp_res]

            # calculate geometric mean of metrics for each peak

            g_mean = [(dist_res[i] + amp_res[i] + pos_res[i]) / 3 for i in range(0, len(amp_res))]

            # compare the residuals and find the minimum

            minres = np.argmin(g_mean)

            # append the closest peaks to the vector

            fit_s_peaks, amp_vector, fit_s_y = new_first_order_peak(picked_peaks_ppm[srange][minres], Jv[ind1],
                                                                    np.arange(len(x_data)), 0.1, uc, 1)

            diff_matrix = np.zeros((len(fit_s_peaks), len(picked_peaks)))

            for i, f in enumerate(fit_s_peaks):

                for j, g in enumerate(picked_peaks):
                    diff_matrix[i, j] = abs(f - g)

            # minimise these distances

            vertical_ind, horizontal_ind = optimise(diff_matrix)

            closest_peaks = np.sort(picked_peaks[horizontal_ind])

            for peak in closest_peaks:
                ind3 = np.abs(picked_peaks - peak).argmin()

                peaks_to_remove.append(picked_peaks[ind3])

                differences.append(picked_peaks_ppm[ind3] - uc.ppm(peak))

    # find the region this peak is in and append it to the list

    for peak in peaks_to_remove:

        for ind2, region in enumerate(peak_regions):

            if (peak > region[0]) & (peak < region[-1]):
                solvent_regions.append(ind2)
                break

    # now remove the selected peaks from the picked peaks list and grouped peaks

    w = np.searchsorted(picked_peaks, peaks_to_remove)

    picked_peaks = np.delete(picked_peaks, w)

    for ind4, peak in enumerate(peaks_to_remove):
        grouped_peaks[solvent_regions[ind4]] = np.delete(grouped_peaks[solvent_regions[ind4]],
                                                         np.where(grouped_peaks[solvent_regions[ind4]] == peak))

    # resimulate the solvent regions

    solvent_region_ind = sorted(list(set(solvent_regions)))

    # now need to reference the spectrum

    # differences = list of differences in ppm found_solvent_peaks - expected_solvent_peaks

    s_differences = sum(differences)

    x_data = x_data + s_differences

    return peak_regions, picked_peaks, grouped_peaks, x_data, solvent_region_ind


def integrate_regions(peak_regions, y_data, noise_std):
    integrals = np.zeros(len(peak_regions))

    for index, region in enumerate(peak_regions):
        w = np.where(y_data[region] > 3 * noise_std)

        integrals[index] = np.sum(y_data[region][w])

    return integrals


def integrate_sim_regions(sim_regions, grouped_peaks, peak_regions, y_data, params, solvent_region_ind):
    sim_integrals = []

    for r, group in enumerate(grouped_peaks):

        region_integral = 0

        for peak in group:
            region_integral += params['A' + str(peak)] * 0.25 * np.pi * params['std' + str(peak)] * (
                        (3 ** 0.5 - 2) * params['vregion' + str(r)] + 2)

        sim_integrals.append(region_integral)

    sim_integrals = np.array(sim_integrals)

    y_integral = np.sum(y_data)

    sim_integral = np.sum(sim_integrals)

    k = sim_integral / y_integral

    integrals = []

    for region in peak_regions:
        integrals.append(np.sum(y_data[region]))

    k = np.array(sim_integrals) / np.array(integrals)

    simr_regions = []

    for region in sim_regions:
        simr_regions.append(np.sum(region))

    integrals = k * integrals

    return integrals


def integral_add(sim_regions, proton_guess):

    integrals = [np.sum(sim) for sim in sim_regions]

    i_sum = np.sum(integrals)

    Inorm = proton_guess / i_sum

    integrals_copy = list(integrals)

    integrals_copy = [0] + integrals_copy

    integral_sum = np.cumsum(np.asarray(integrals_copy))
    integral_sum = integral_sum / np.sum(integrals_copy)

    cummulative_vectors = []

    t_sum = 0
    for region in sim_regions:
        t_sum += np.sum(region)
        cummulative_vectors.append(np.cumsum(region))

    for i, region in enumerate(cummulative_vectors):
        cummulative_vectors[i] = region / t_sum

    return integral_sum, cummulative_vectors


def normalise_integration(integrals, initial_proton_guess):
    i_sum = np.sum(integrals)

    integrals = integrals / i_sum

    norm_integrals = integrals * initial_proton_guess

    return norm_integrals


def remove_impurities(integrals, peak_regions, grouped_peaks, picked_peaks, sim_regions):
    # find rounded values

    r = sum_round(integrals)

    r = np.array(r)

    to_remove = np.where(r < 0.5)[0]

    number_of_impurities = len(to_remove)

    peaks_to_remove = []

    for group in to_remove:
        peaks_to_remove.append(grouped_peaks[group])

    whdel = np.where(picked_peaks == peaks_to_remove)

    picked_peaks = np.delete(picked_peaks, whdel)

    integrals = np.delete(integrals, to_remove)

    peak_regions = np.delete(peak_regions, to_remove)

    grouped_peaks = np.delete(grouped_peaks, to_remove)

    sim_regions = np.delete(sim_regions, to_remove)

    rounded_integrals = r[r > 0.5]

    return grouped_peaks, integrals, peak_regions, picked_peaks, number_of_impurities, sim_regions,rounded_integrals


def proton_count(file):
    sdf_file = open(str(file), 'r')

    txt = sdf_file.readlines()

    counter = 0
    for line in txt:
        if ' H   ' in line:
            counter += 1

    return counter


def integral_score(integrals, structure_protons, proton_guess, l_protons, impurities):

    r_int = sum_round(integrals)

    r_int = np.array(r_int)

    sum_r = np.sum(r_int)

    new_integrals = []

    for ind, integral in enumerate(integrals):
        new_integrals += [integral] * int(r_int[ind])

    new_integrals = np.array(new_integrals)

    differences = new_integrals % 0.5

    std = structure_protons / 8

    diff = abs(proton_guess - structure_protons)

    probs = 4 * (1 - norm.cdf(differences, loc=0, scale=1 / 16))

    mean = gmean(probs) * (1 - norm.cdf(diff, loc=0, scale=std)) * (1 / 2 ** impurities)

    # only allow intergrals that are between the expected number and that number - the number of labile protons

    if (sum_r < structure_protons - l_protons):
        mean = 0

    return mean


def labile_protons(file):

    mol = Chem.MolFromMolFile(file, removeHs=False)
    count = 0

    for atom in mol.GetAtoms():
        if atom.GetSymbol() == 'O':
            for neighbor in atom.GetNeighbors():
                if neighbor.GetSymbol() == 'H':
                    count +=1

    return count


def find_integrals(file, peak_regions, grouped_peaks, sim_regions,
                   picked_peaks, params, y_data, solvent_region_ind):
    # count number of protons in the file

    structure_protons = proton_count(str(file).split('.sdf')[0] + ".sdf")

    # find number of methyl groups

    m = methyl_protons(str(file).split('.sdf')[0] + ".sdf")

    number_of_methyl_groups = len(m)

    print('     number of protons = ' + str(structure_protons))

    # count the number of labile protons in the structure

    l_protons = labile_protons(file)

    print('     number of labile protons = ' + str(l_protons))

    count = 0

    # allow guesses of number of protons in the spectrum between: structure_protons - l_protons, 2 * structure_protons

    number_vector = np.arange(structure_protons - l_protons, 2 * structure_protons)

    scores = np.zeros(len(number_vector))

    integrals = integrate_sim_regions(sim_regions, grouped_peaks, peak_regions, y_data, params, solvent_region_ind)

    for proton_guess in number_vector:

        norm_integrals = normalise_integration(integrals, proton_guess)

        # remove impurities

        # find number of impurities removed

        # find which should be classed as impurities

        r = sum_round(norm_integrals)

        r = np.array(r)

        impurities = len(r[r < 1])

        norm_integrals = norm_integrals[r > 0.5]

        scores[count] = integral_score(norm_integrals, structure_protons, proton_guess, l_protons, impurities)

        number_of_methyl_groups_integral = np.sum((r - (r % 3)) // 3)

        if number_of_methyl_groups_integral < number_of_methyl_groups:
            scores[count] = 0

        count += 1

    wh = np.argmax(scores)

    best_fit = number_vector[wh]

    print("the best fit number of protons is " + str(best_fit))

    # normalise using this number

    integrals = normalise_integration(integrals, best_fit)

    grouped_peaks, integrals, peak_regions, picked_peaks_, impurities, sim_regions,rounded_integrals = remove_impurities(integrals,
                                                                                                       peak_regions,
                                                                                                       grouped_peaks,
                                                                                                       picked_peaks,
                                                                                                       sim_regions)

    integral_sum, cummulative_vectors = integral_add(sim_regions, best_fit)

    total_r = 0

    total = 0

    for i in range(0, len(integrals)):
        total += integrals[i]
        total_r += rounded_integrals[i]

    return peak_regions, grouped_peaks, sim_regions, integral_sum, cummulative_vectors, rounded_integrals, structure_protons, \
           best_fit, total_r


def sum_round(a):
    rounded = [round(i, 0) for i in a]

    error = sum(a) - sum(rounded)

    n = int(round(error / 1))

    new = rounded[:]

    for _, i in sorted(((a[i] - rounded[i], i) for i in range(len(a))), reverse=n > 0)[:abs(n)]:
        new[i] += n / abs(n)

    return new


def weighted_region_centres(peak_regions, total_spectral_ydata):
    centres = []

    for region in peak_regions:
        w = total_spectral_ydata[region] ** 2

        wx = region * w

        xbar = np.sum(wx) / np.sum(total_spectral_ydata[region] ** 2)

        centres.append(int(xbar))

    return centres


def methyl_protons(file):

    mol = Chem.MolFromMolFile(file)

    methyl_protons = []

    for atom in mol.GetAtoms():
        if atom.GetSymbol() == 'C':
            nbrprotons = []
            for nbr in atom.GetNeighbors():
                if nbr.GetSymbol() == 'H':
                    nbrprotons.append(f'H{nbr.GetIdx()+1}')
            if len(nbrprotons) == 3:
                methyl_protons.append(nbrprotons)


    return methyl_protons




import logging
import os
import shutil
import sys
import time
import pathlib
import subprocess
import re

from dp5.dft import BaseDFTMethod

logger = logging.getLogger(__name__)


class DFTMethod(BaseDFTMethod):
    def __init__(self, settings):
        super().__init__(settings)

        self.tag = "g"
        self.input_format = ".com"
        self.output_format = ".out"

        if "GAUS_EXEDIR" in os.environ:
            gausdir = os.environ["GAUSS_EXEDIR"]
            if shutil.which(os.path.join(gausdir, "g09")) is None:
                self.executable = os.path.join(gausdir, "g16")
            else:
                self.executable = os.path.join(gausdir, "g09")
        else:
            self.executable = self.settings["executable"]["gaussian"]

        if shutil.which(self.executable) is None:
            logger.warning(
                f"Gaussian executable not found at {self.executable}. Calculations will not run!"
            )

    def __repr__(self):
        return "Gaussian"

    def write_file(self, filename, coordinates, atoms, charge, type):

        with open(f"{filename}.com", "w") as f:

            if self.settings["num_processors"] > 1:
                f.write(f"%nprocshared={self.settings['num_processors']}\n")

            f.write(f"%mem={self.settings['memory']}MB\n" + f"%chk={filename}.chk\n")

            if type == "nmr":
                f.write(self.nmr_options())
            elif type == "e":
                f.write(self.e_options())
            elif type == "opt":
                f.write(self.opt_options())

            f.write(f"\n{filename}\n\n" + f"{charge} 1\n")

            for atom, (x, y, z) in zip(atoms, coordinates):
                f.write(f"{atom}  {x:.6f}  {y:.6f}  {z:.6f}\n")

            f.write("\n")

            # type specific footer

    def nmr_options(self):

        ultrafine = (
            " int=ultrafine"
            if (self.settings["n_functional"]).lower() == "m062x"
            else " "
        )
        solvent = (
            f" scrf=(solvent={self.settings['solvent']})"
            if self.settings["solvent"]
            else " "
        )

        route = (
            f"# {self.settings['n_functional']}/{self.settings['n_basis_set']}"
            f"{ultrafine} nmr=giao{solvent}\n"
        )

        return route

    def e_options(self):

        ultrafine = (
            " int=ultrafine"
            if (self.settings["e_functional"]).lower() == "m062x"
            else " "
        )
        solvent = (
            f" scrf=(solvent={self.settings['solvent']})"
            if self.settings["solvent"]
            else " "
        )

        route = (
            f"# {self.settings['e_functional']}/{self.settings['e_basis_set']}"
            f"{ultrafine}{solvent}\n"
        )

        return route

    def opt_options(self):

        ultrafine = (
            " int=ultrafine"
            if (self.settings["o_functional"]).lower() == "m062x"
            else " "
        )
        solvent = (
            f" scrf=(solvent={self.settings['solvent']})"
            if self.settings["solvent"]
            else " "
        )
        calc_fc = ",CalcFC" if self.settings["calc_force_constants"] else ""
        step_size = (
            f",MaxStep={100*self.settings['opt_step_size']}"
            if self.settings["opt_step_size"] != 0.3
            else ""
        )

        route = (
            f"# {self.settings['o_functional']}/{self.settings['o_basis_set']}"
            f"{ultrafine} Opt=(maxcycles={self.settings['max_opt_cycles']}{calc_fc}){step_size}{solvent}\n"
        )

        return route

    def _run_calcs(self, jobs: list):
        return super()._run_calcs(jobs)

    def prepare_command(self, file):
        return f"{self.executable} < {file}{self.input_format} > {file}{self.output_format}"

    def read_file(self, file):

        atoms = []
        coordinates = []
        energy = None
        shieldings = []
        shielding_labels = []
        completed = False
        opt_converged = False

        with open(file, "r") as f:
            while not completed:
                try:
                    line = next(f)
                    # reads geometry
                    if (
                        line.strip() == "Standard orientation:"
                        or line.strip() == "Input orientation:"
                    ):
                        atoms = []
                        coordinates = []
                        # skips lines
                        for _ in range(5):
                            line = next(f)

                        while not "----------" in line:
                            atom_idx, atom_number, _, *coords = line.split()
                            atom_label = self.atom_num_to_symbol(int(atom_number))
                            atoms.append(atom_label)
                            coordinates.append([float(x) for x in coords])
                            line = next(f)

                    # reads energy
                    if "SCF Done:" in line:
                        start = line.index(") =")
                        end = line.index("A.U.")
                        energy = float(line[start + 4 : end]) * self.hartree_to_kJ

                    # reads shieldings
                    if line == " SCF GIAO Magnetic shielding tensor (ppm):\n":
                        shieldings = []
                        line = next(f)
                        while "Isotropic" in line:
                            atom_idx, atom_label, _, _, shielding, *_ = line.split()
                            shieldings.append(float(shielding))
                            shielding_labels.append(f"{atom_label}{atom_idx}")

                            for _ in range(5):
                                line = next(f)

                    if "Stationary point found" in line:
                        opt_converged = True

                    if (
                        line == "Error termination request processed by link 9999."
                        or "Normal termination" in line
                    ):
                        completed = True

                except StopIteration:
                    logger.error(
                        f"Calculations in {file} have not terminated normally."
                    )
                    break

        return (
            atoms,
            coordinates,
            energy,
            shieldings,
            shielding_labels,
            completed,
            opt_converged,
        )



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 12 12:52:21 2015

@author: ke291

Contains all of the Tinker specific code for input generation, calculation
execution and output interpretation. Called by PyDP4.py.
"""

import os
import shlex
import shutil
import subprocess
import logging

from rdkit import Chem
from rdkit.Chem.rdMolHash import MolHash, HashFunction


from dp5.conformer_search.base_cs_method import BaseConfSearch, ConfData
import dp5.conformer_search.sdftinkerxyzpy as sdftinkerxyzpy


logger = logging.getLogger(__name__)
__all__ = ["ConfSearchMethod"]


class ConfSearchMethod(BaseConfSearch):

    def __init__(self, settings):
        self.settings = settings
        if not self.settings["executable"]["tinker"]:
            logger.warning(
                "No Tinker executable specified. Please check your config file."
            )
        self.executable = self.settings["executable"]["tinker"]

        self.tinker_exe = os.path.join(self.executable, "bin", "scan")
        self.ff_params = os.path.join(self.executable, "params", "mmff.prm")

    def __repr__(self):
        return "Tinker"

    def prepare_input(self, inputs):
        for input in inputs:
            convinp = sdftinkerxyzpy.main(input)
            outp = subprocess.check_output(convinp + input + ".sdf", shell=True)
        return inputs

    def _run(self):

        completed = 0
        total = len(self.inputs)
        outputs = []

        if shutil.which(self.tinker_exe) is None:
            logger.critical(f"Could not find Tinker executable at {self.tinker_exe}")
            raise RuntimeError(f"Tinker executable not found at {self.tinker_exe}")

        if not os.path.exists(self.ff_params):
            logger.critical(f"Could not find MMFF parameters at {self.ff_params}")
            raise RuntimeError(f"MMFF parameters not found at {self.ff_params}")

        for input in self.inputs:
            if os.path.exists(f"{input}.tout") and os.path.exists(f"{input}.arc"):
                logger.info(f"Found output files for {input}")

            else:
                logger.info(f"Running conformational search for {input}")
                with open(f"{input}.tout", "w") as tout:
                    subprocess.run(
                        f"{self.tinker_exe} {input}.xyz {self.ff_params} 0 10 20 0.00001",
                        text=True,
                        shell=True,
                        check=True,
                        stdout=tout,
                    )

            outputs.append(input)
            completed = completed + 1
            logger.info(f"{completed} out of {total} conformer searches complete")

        return outputs

    def parse_output(self):
        final_list = []

        for file in self.inputs:
            final_list.append(self._parse_output(file))

        return final_list

    def _parse_output(self, file):

        all_energies, charge = self._get_energies_charge(file)
        atoms, all_conformers = self._read_arc(file)

        conformers = []
        energies = []

        for energy, conformer in zip(all_energies, all_conformers):
            if energy < min(energies) + self.settings["energy_cutoff"]:
                energies.append(energy)
                conformers.append(conformer)

        conf_data = ConfData(atoms, conformers, charge, energies)
        return conf_data

    def _get_energies_charge(self, file):

        with open(f"{file}.tout", "r") as f:
            inp = f.readlines()

        if len(inp) < 13:
            logger.critical(f"{file}.tout is incomplete")
            raise ValueError(f"{file}.tout is incomplete")

        energies = []

        # Get the conformer energies from the file
        energies = []
        for line in inp[13:]:
            data = line[:-1].split("  ")
            data = [_f for _f in data if _f]
            if len(data) >= 3:
                if "Map" in data[0] and "Minimum" in data[1]:
                    energies.append(float(data[-1]))

        mol = Chem.MolFromMolFile(f"{file}.sdf", removeHs=False)
        charge = int(MolHash(mol, HashFunction.NetCharge))

        return energies, charge

    def _read_arc(self, file):

        def GetAtomSymbol(AtomNum):
            Lookup = [
                "H",
                "He",
                "Li",
                "Be",
                "B",
                "C",
                "N",
                "O",
                "F",
                "Ne",
                "Na",
                "Mg",
                "Al",
                "Si",
                "P",
                "S",
                "Cl",
                "Ar",
                "K",
                "Ca",
                "Sc",
                "Ti",
                "V",
                "Cr",
                "Mn",
                "Fe",
                "Co",
                "Ni",
                "Cu",
                "Zn",
                "Ga",
                "Ge",
                "As",
                "Se",
                "Br",
                "Kr",
                "Rb",
                "Sr",
                "Y",
                "Zr",
                "Nb",
                "Mo",
                "Tc",
                "Ru",
                "Rh",
                "Pd",
                "Ag",
                "Cd",
                "In",
                "Sn",
                "Sb",
                "Te",
                "I",
                "Xe",
                "Cs",
                "Ba",
                "La",
                "Ce",
                "Pr",
                "Nd",
                "Pm",
                "Sm",
                "Eu",
                "Gd",
                "Tb",
                "Dy",
                "Ho",
                "Er",
                "Tm",
                "Yb",
                "Lu",
                "Hf",
                "Ta",
                "W",
                "Re",
                "Os",
                "Ir",
                "Pt",
                "Au",
                "Hg",
                "Tl",
                "Pb",
                "Bi",
                "Po",
                "At",
                "Rn",
            ]

            # RK: original implementation is maintained
            if AtomNum > 0 and AtomNum < len(Lookup):
                return Lookup[AtomNum - 1]
            else:
                logger.error(
                    f"Element with atomic number {AtomNum} not supported, will return 0"
                )
                return 0

        atypes, anums = self._extract_atom_types()

        with open(f"{file}.arc", "r") as conffile:
            confdata = conffile.readlines()

        # output data: conformers - list of x,y,z lists, atoms - list of atoms
        conformers = []
        atoms = []
        atypes = [x[:3] for x in atypes]

        for line in confdata:
            data = [_f for _f in line.split("  ") if _f]
            if len(data) < 3:
                conformers.append([])
            else:
                if len(conformers) == 1:
                    anum = anums[atypes.index(data[1][:3])]
                    atoms.append(GetAtomSymbol(anum))
                conformers[-1].append([x for x in data[2:5]])

        return atoms, conformers

    def _extract_atom_types(self):
        atomtypes = []
        atomnums = []

        with open(self.ff_params, "w") as f:
            for line in f:
                if line.split(" ")[0] == "atom":
                    data = shlex.split(line, posix=False)
                    atomtypes.append(data[3])
                    atomnums.append(int(data[-3]))
        return atomtypes, atomnums



import logging

import numpy as np
from lmfit import Minimizer, Parameters
from scipy.interpolate import InterpolatedUnivariateSpline

from dp5.nmr_processing.helper_functions import normalise_intensities


logger = logging.getLogger(__name__)


def spectral_processing(total_spectral_ydata):
    """Arguments:
    - ydata: complex array with frequency-domain FID data
    Returns:
    - tydata (np.array): phased and baseline-corrected frequency data
    – corr_distance (int): autocorrelation distance
    – sigma (float): standard deviation of noise
    – peak_regions(np.array[np.array]): regoins containing peak signals
    """
    logger.info("Processing Proton Spectrum")

    corr_distance = estimate_autocorrelation(total_spectral_ydata)
    # baseline and phasing
    tydata = ACMEWLRhybrid(total_spectral_ydata, corr_distance)

    # find final noise distribution
    classification, sigma = baseline_find_signal(
        tydata, corr_distance, True, 1)

    # fall back phasing if fit doesnt converge
    # calculate negative area
    # draw regions

    c1 = np.roll(classification, 1)
    diff = classification - c1
    s_start = np.where(diff == 1)[0]
    s_end = np.where(diff == -1)[0] - 1

    # region start and end indices are known
    # indices inbetween are filled in
    peak_regions = np.empty(shape=s_start.shape[0], dtype=object)
    peak_regions[:] = [np.arange(s, e) for s, e in zip(s_start, s_end)]

    tydata = normalise_intensities(tydata)

    return tydata, corr_distance, sigma, peak_regions


def estimate_autocorrelation(total_spectral_ydata):
    # note this region may have a baseline distortion

    y = np.real(total_spectral_ydata[0:10000])

    params = Parameters()

    # define a basleine polynomial

    order = 6

    for p in range(order + 1):
        params.add("p" + str(p), value=0)

    def poly(params, order, y):

        bl = np.zeros(len(y))
        x = np.arange(len(y))

        for p in range(order + 1):
            bl += params["p" + str(p)] * x ** (p)

        return bl

    def res(params, order, y):

        bl = poly(params, order, y)

        r = abs(y - bl)

        return r

    out = Minimizer(res, params, fcn_args=(order, y))

    results = out.minimize()

    bl = poly(results.params, order, y)

    y = y - bl

    t0 = np.sum(y * y)

    c = 1

    tc = 1

    t = []

    while tc > 0.36:
        tc = np.sum(np.roll(y, c) * y) / t0

        t.append(tc)

        c += 1

    return c


def ACMEWLRhybrid(y, corr_distance):
    """Hybrid ACME-Weighted Linear Regression phase correction algorithm"""

    def residual_function(params, im, real):

        # phase the region

        data = ps(params, im, real, 0)

        # make new baseline for this region

        r = np.linspace(data[0], data[-1], len(real))

        # find negative area

        data -= r

        ds1 = np.abs((data[1:] - data[:-1]))

        p1 = ds1 / np.sum(ds1)

        # Calculation of entropy
        p1[p1 == 0] = 1

        h1 = -p1 * np.log(p1)
        h1s = np.sum(h1)

        # Calculation of penalty
        pfun = 0.0

        as_ = data - np.abs(data)

        sumas = np.sum(as_)

        if sumas < 0:
            pfun = (as_[1:] / 2) ** 2

        p = np.sum(pfun)

        return h1s + 1000 * p

    # find regions

    classification, sigma = baseline_find_signal(y, corr_distance, True, 1)

    c1 = np.roll(classification, 1)

    diff = classification - c1

    s_start = np.where(diff == 1)[0]

    s_end = np.where(diff == -1)[0] - 1

    peak_regions = []

    for r in range(len(s_start)):
        peak_regions.append(np.arange(s_start[r], s_end[r]))

    # for region in peak_regions:
    #    plt.plot(region,y[region],color = 'C1')

    # phase each region independently

    phase_angles = []

    weights = []

    centres = []

    for region in peak_regions:
        params = Parameters()

        params.add("p0", value=0, min=-np.pi, max=np.pi)

        out = Minimizer(
            residual_function, params, fcn_args=(
                np.imag(y[region]), np.real(y[region]))
        )

        results = out.minimize("brute")

        p = results.params

        phase_angles.append(p["p0"] * 1)

        # find weight

        data = ps(p, np.imag(y[region]), np.real(y[region]), 0)

        # make new baseline for this region

        r = np.linspace(data[0], data[-1], len(data))

        # find negative area

        res = data - r

        weights.append(abs(np.sum(res[res > 0] / np.sum(y[y > 0]))))

        centres.append(np.median(region) / len(y))

    sw = sum(weights)

    weights = [w / sw for w in weights]

    # do weighted linear regression on the regions

    # do outlier analysis

    switch = 0

    centres = np.array(centres)

    weights = np.array(weights)

    sweights = np.argsort(weights)[::-1]

    phase_angles = np.array(phase_angles)

    ind1 = 0

    while switch == 0:

        intercept, gradient = np.polynomial.polynomial.polyfit(
            centres, phase_angles, deg=1, w=weights
        )

        predicted_angles = gradient * centres + intercept

        weighted_res = np.abs(predicted_angles - phase_angles) * weights

        # find where largest weighted residual is

        max_res = sweights[ind1]

        s = 0

        if phase_angles[max_res] > 0:

            s = -1

            phase_angles[max_res] -= 2 * np.pi

        else:

            s = +1

            phase_angles[max_res] += 2 * np.pi

        intercept1, gradient1 = np.polynomial.polynomial.polyfit(
            centres, phase_angles, deg=1, w=weights
        )

        new_predicted_angles = gradient1 * centres + intercept1

        new_weighted_res = np.abs(
            new_predicted_angles - phase_angles) * weights

        if np.sum(new_weighted_res) > np.sum(weighted_res):

            switch = 1

            phase_angles[max_res] += -2 * np.pi * s

        ind1 += 1

    # phase the data

    p_final = Parameters()

    p_final.add("p0", value=intercept)
    p_final.add("p1", value=gradient)

    y = ps(p_final, np.imag(y), np.real(y), 1)

    classification, sigma = baseline_find_signal(y, corr_distance, True, 1)
    r = gen_baseline(np.real(y), classification, corr_distance)
    y -= r

    return np.real(y)


def gen_baseline(y_data, sn_vector, corr_distance):
    points = np.arange(len(y_data))

    spl = InterpolatedUnivariateSpline(
        points[sn_vector == 0], y_data[sn_vector == 0], k=1
    )

    r = spl(points)

    # is corr distance odd or even

    if corr_distance % 2 == 0:
        kernel = np.ones((corr_distance + 1) * 10) / ((corr_distance + 1) * 10)
    else:
        kernel = np.ones((corr_distance) * 10) / ((corr_distance) * 10)

    r = np.convolve(r, kernel, mode="same")

    return r


def ps(param, im, real, phase_order):
    x = np.linspace(0, 1, len(real))

    angle = np.zeros(len(x))

    for p in range(phase_order + 1):
        angle += param["p" + str(p)] * x ** (p)

    # phase the data

    R = real * np.cos(angle) - im * np.sin(angle)

    return R


def baseline_find_signal(y_data, cdist, dev, t):
    wd = int(cdist) * 10

    sd_all = _get_sd(y_data, wd)

    snvectort = np.zeros(len(y_data))

    sv = []

    for i in range(0, 4 * cdist):
        x = np.arange(i + wd, len(y_data) - wd, 4 * cdist)

        sample = y_data[x]

        sd_set = _get_sd(sample, wd)

        s = _find_noise_sd(sd_set, 0.999)

        sv.append(s)

    sigma = np.mean(sv)

    b = np.linspace(-0.001, 0.001, 1000)

    if dev == True:
        w = np.where(sd_all > t * sigma)[0]

    else:
        w = np.where(y_data > t * sigma)[0]

    snvectort[w] = 1

    sn_vector = np.zeros(len(y_data))

    w = cdist

    for i in np.arange(len(sn_vector)):
        if snvectort[i] == 1:
            sn_vector[np.maximum(0, i - w): np.minimum(i + w, len(sn_vector))] = 1

    return sn_vector, sigma


def _rolling_window(a, window):
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


def _get_sd(data, k):
    return np.std(_rolling_window(data, k), -1)


def _find_noise_sd(sd_set, ratio):
    """Calculate the median m1 from SDset. exclude the elements greater
    than 2m1from SDset and recalculate the median m2. Repeat until
    m2/m1 converge(sd_set)"""
    m1 = np.median(sd_set)
    S = sd_set <= 2.0 * m1
    tmp = S * sd_set
    sd_set = tmp[tmp != 0]
    m2 = np.median(sd_set)
    while m2 / m1 < ratio:
        m1 = np.median(sd_set)
        S = sd_set <= 2.0 * m1
        tmp = S * sd_set
        sd_set = tmp[tmp != 0]

    return m2



"""Collection of helper file for DP4 and DP5 analysis"""

import pickle
from pathlib import Path

import numpy as np
from scipy.stats import linregress


def _scale_nmr(calc_shifts, exp_shifts):
    """
    Linear correction to calculated shifts
    Arguments:
    - calc_shifts(list): calculated shifts
    - exp_shifts(list): experimental shifts as assigned

    Returns:
    - scaled calculated shifts for error analysis
    """

    if len(calc_shifts) > 1:
        slope, intercept, r_value, p_value, std_err = linregress(
            exp_shifts, calc_shifts
        )
        if not isinstance(calc_shifts, np.ndarray) or isinstance(
            exp_shifts, np.ndarray
        ):
            calc_shifts = np.array(calc_shifts)
            exp_shifts = np.array(exp_shifts)
        intercept = np.nan_to_num(intercept, nan=0)
        slope = np.nan_to_num(slope, nan=1)
        scaled_shifts = (calc_shifts - intercept) / slope
    else:
        scaled_shifts = calc_shifts

    return scaled_shifts


def scale_nmr(calc_shifts, exp_shifts):
    """
    Linear correction to calculated shifts for a conformer.

    For each geometry, applies a linear correction to the shifts, returns sclaed shifts. Can later be used for caclucating corrected errors.

    Arguments:
    - calc_shifts(list): calculated shifts
    - exp_shifts(list): experimental shifts as assigned

    Returns:
    - scaled calculated shifts for error analysis
    """
    if not isinstance(calc_shifts, np.ndarray) or not isinstance(
        exp_shifts, np.ndarray
    ):
        calc_shifts = np.array(calc_shifts)
        exp_shifts = np.array(exp_shifts)

    if len(calc_shifts.shape) > 1:
        scale_axis = lambda conf_shifts: _scale_nmr(conf_shifts, exp_shifts)
        scaled_shifts = np.apply_along_axis(
            scale_axis, 1, np.array(calc_shifts, dtype=np.float32)
        )
    else:
        scaled_shifts = _scale_nmr(calc_shifts, exp_shifts)

    return scaled_shifts


class AnalysisData:
    """Container class for DP4 or DP5 analysis."""

    # TODO: incorporate json
    def __init__(self, mols, path):
        """
        Args:
            mols: takes Molecule objects used in the analysis
            path: path to load/save the analysis data
        """
        self.mols = [mol.input_file for mol in mols]
        self.path = path

    @property
    def exists(self):
        """Checks if a file exists at the provided path"""
        return Path(self.path).exists()

    @property
    def values_dict(self):
        keys_to_exclude = ("path", "exists")
        return {k: v for k, v in self.__dict__.items() if k not in keys_to_exclude}

    def load(self):
        with open(self.path, "rb") as f:
            data = pickle.load(f)
            for key, value in data.items():
                setattr(self, key, value)

    def __iter__(self):
        return (
            dict(zip(self.values_dict.keys(), values))
            for values in zip(*self.values_dict.values())
        )

    def save(self):
        with open(self.path, "wb") as f:
            pickle.dump(self.values_dict, f)

    @property
    def by_mol(self):
        """Returns rearranged dictionary of lists (grouped by a property) into list of dictionaries (grouped by a molecule)"""

        return [
            dict(zip(self.values_dict.keys(), values))
            for values in zip(*self.values_dict.values())
        ]

    def append(self, mol_dict):
        """Adds data from a molecule.

        Searches for relevant attributes, appends the data to the class"""
        for key, value in mol_dict.items():
            if key not in self.__dict__.keys():
                self.__dict__[key] = [value]
            else:
                self.__dict__[key].append(value)

    def from_mol_dicts(self, dicts):
        """Reads molecular dictionaries containing properties, appends them as class attributes"""
        transposed_dict = dict()
        for dp4d in dicts:
            for key, value in dp4d.items():
                if key in transposed_dict.keys():
                    transposed_dict[key].append(value)
                else:
                    transposed_dict[key] = [value]
        self.__dict__.update(transposed_dict)



import numpy as np

from sklearn.preprocessing import RobustScaler
from scipy import sparse
from sklearn.utils import check_array
from sklearn.utils.validation import FLOAT_DTYPES
from sklearn.preprocessing._data import _handle_zeros_in_scale

class RobustNanScaler(RobustScaler):
    
    def _check_array(self, X, copy):
        """Makes sure centering is not enabled for sparse matrices."""
        X = check_array(X, accept_sparse=('csr', 'csc'), copy=self.copy,
                        estimator=self, dtype=FLOAT_DTYPES, force_all_finite=False)

        if sparse.issparse(X):
            if self.with_centering:
                raise ValueError(
                    "Cannot center sparse matrices: use `with_centering=False`"
                    " instead. See docstring for motivation and alternatives.")
        return X

    
    def fit(self, X, y=None):
        if sparse.issparse(X):
            raise TypeError("RobustScaler cannot be fitted on sparse inputs")
        X = self._check_array(X, self.copy)
        if self.with_centering:
            self.center_ = np.nanmedian(X, axis=0)

        if self.with_scaling:
            q_min, q_max = self.quantile_range
            if not 0 <= q_min <= q_max <= 100:
                raise ValueError("Invalid quantile range: %s" %
                                 str(self.quantile_range))

            q = np.nanpercentile(X, self.quantile_range, axis=0)
            self.scale_ = (q[1] - q[0])
            self.scale_ = _handle_zeros_in_scale(self.scale_, copy=False)
        return self



import logging, sys

import numpy as np
from tqdm import tqdm
from scipy.linalg import eigh

from rdkit import Chem
from rdkit.Chem import MolFromSmiles, MolToSmiles, AddHs

from . import features
from .features import Tokenizer
import time


class SmilesPreprocessor(object):
    """ Given a list of SMILES strings, encode these molecules as atom and
    connectivity feature matricies.

    Example:
    >>> preprocessor = SmilesPreprocessor(explicit_hs=False)
    >>> inputs = preprocessor.fit(data.smiles)
    """

    def __init__(self, explicit_hs=True, atom_features=None, bond_features=None):
        """

        explicit_hs : bool
            whether to tell RDkit to add H's to a molecule.
        atom_features : function
            A function applied to an rdkit.Atom that returns some
            representation (i.e., string, integer) for the Tokenizer class.
        bond_features : function
            A function applied to an rdkit Bond to return some description.

        """

        self.atom_tokenizer = Tokenizer()
        self.bond_tokenizer = Tokenizer()
        self.explicit_hs = explicit_hs

        if atom_features is None:
            atom_features = features.atom_features

        if bond_features is None:
            bond_features = features.bond_features_v1

        self.atom_features = atom_features
        self.bond_features = bond_features


    def fit(self, smiles_iterator):
        """ Fit an iterator of SMILES strings, creating new atom and bond
        tokens for unseen molecules. Returns a dictionary with 'atom' and
        'connectivity' entries """
        return list(self.preprocess(smiles_iterator, train=True))


    def predict(self, smiles_iterator):
        """ Uses previously determined atom and bond tokens to convert a SMILES
        iterator into 'atom' and 'connectivity' matrices. Ensures that atom and
        bond classes commute with previously determined results. """
        return list(self.preprocess(smiles_iterator, train=False))


    def preprocess(self, smiles_iterator, train=True):

        self.atom_tokenizer.train = train
        self.bond_tokenizer.train = train

        for smiles in tqdm(smiles_iterator):
            yield self.construct_feature_matrices(smiles)


    @property
    def atom_classes(self):
        """ The number of atom types found (includes the 0 null-atom type) """
        return self.atom_tokenizer.num_classes + 1


    @property
    def bond_classes(self):
        """ The number of bond types found (includes the 0 null-bond type) """
        return self.bond_tokenizer.num_classes + 1


    def construct_feature_matrices(self, smiles):
        """ construct a molecule from the given smiles string and return atom
        and bond classes.

        Returns
        dict with entries
        'n_atom' : number of atoms in the molecule
        'n_bond' : number of bonds in the molecule
        'atom' : (n_atom,) length list of atom classes
        'bond' : (n_bond,) list of bond classes
        'connectivity' : (n_bond, 2) array of source atom, target atom pairs.

        """

        mol = MolFromSmiles(smiles)
        if self.explicit_hs:
            mol = AddHs(mol)

        n_atom = len(mol.GetAtoms())
        n_bond = 2 * len(mol.GetBonds())

        # If its an isolated atom, add a self-link
        if n_bond == 0:
            n_bond = 1

        atom_feature_matrix = np.zeros(n_atom, dtype='int')
        bond_feature_matrix = np.zeros(n_bond, dtype='int')
        connectivity = np.zeros((n_bond, 2), dtype='int')

        bond_index = 0

        atom_seq = mol.GetAtoms()
        atoms = [atom_seq[i] for i in range(n_atom)]

        for n, atom in enumerate(atoms):

            # Atom Classes
            atom_feature_matrix[n] = self.atom_tokenizer(
                self.atom_features(atom))

            start_index = atom.GetIdx()

            for bond in atom.GetBonds():
                # Is the bond pointing at the target atom
                rev = bond.GetBeginAtomIdx() != start_index

                # Bond Classes
                bond_feature_matrix[n] = self.bond_tokenizer(
                    self.bond_features(bond, flipped=rev))

                # Connectivity
                if not rev:  # Original direction
                    connectivity[bond_index, 0] = bond.GetBeginAtomIdx()
                    connectivity[bond_index, 1] = bond.GetEndAtomIdx()

                else:  # Reversed
                    connectivity[bond_index, 0] = bond.GetEndAtomIdx()
                    connectivity[bond_index, 1] = bond.GetBeginAtomIdx()

                bond_index += 1


        return {
            'n_atom': n_atom,
            'n_bond': n_bond,
            'atom': atom_feature_matrix,
            'bond': bond_feature_matrix,
            'connectivity': connectivity,
        }


class ConnectivityAPreprocessor(object):
    """ Given a list of SMILES strings, encode these molecules as atom and
    connectivity feature matricies.

    Example:
    >>> preprocessor = SmilesPreprocessor(explicit_hs=False)
    >>> inputs = preprocessor.fit(data.smiles)
    """

    def __init__(self, explicit_hs=True, atom_features=None, bond_features=None):
        """

        explicit_hs : bool
            whether to tell RDkit to add H's to a molecule.
        atom_features : function
            A function applied to an rdkit.Atom that returns some
            representation (i.e., string, integer) for the Tokenizer class.
        bond_features : function
            A function applied to an rdkit Bond to return some description.

        """

        self.atom_tokenizer = Tokenizer()
        self.bond_tokenizer = Tokenizer()
        self.explicit_hs = explicit_hs

        if atom_features is None:
            atom_features = features.atom_features_v1

        if bond_features is None:
            bond_features = features.bond_features_v1

        self.atom_features = atom_features
        self.bond_features = bond_features


    def fit(self, smiles_iterator):
        """ Fit an iterator of SMILES strings, creating new atom and bond
        tokens for unseen molecules. Returns a dictionary with 'atom' and
        'connectivity' entries """
        return list(self.preprocess(smiles_iterator, train=True))


    def predict(self, smiles_iterator):
        """ Uses previously determined atom and bond tokens to convert a SMILES
        iterator into 'atom' and 'connectivity' matrices. Ensures that atom and
        bond classes commute with previously determined results. """
        return list(self.preprocess(smiles_iterator, train=False))


    def preprocess(self, smiles_iterator, train=True):

        self.atom_tokenizer.train = train
        self.bond_tokenizer.train = train

        for smiles in tqdm(smiles_iterator):
            yield self.construct_feature_matrices(smiles)


    @property
    def atom_classes(self):
        """ The number of atom types found (includes the 0 null-atom type) """
        return self.atom_tokenizer.num_classes + 1


    @property
    def bond_classes(self):
        """ The number of bond types found (includes the 0 null-bond type) """
        return self.bond_tokenizer.num_classes + 1


    def construct_feature_matrices(self, smiles):
        """ construct a molecule from the given smiles string and return atom
        and bond classes.

        Returns
        dict with entries
        'n_atom' : number of atoms in the molecule
        'n_bond' : number of bonds in the molecule
        'atom' : (n_atom,) length list of atom classes
        'bond' : (n_bond,) list of bond classes
        'connectivity' : (n_bond, 2) array of source atom, target atom pairs.

        """

        mol = MolFromSmiles(smiles)
        if self.explicit_hs:
            mol = AddHs(mol)

        n_atom = len(mol.GetAtoms())
        n_bond = 2 * len(mol.GetBonds())

        # If its an isolated atom, add a self-link
        if n_bond == 0:
            n_bond = 1

        atom_feature_matrix = np.zeros(n_atom, dtype='int')
        bond_feature_matrix = np.zeros(n_bond, dtype='int')
        connectivity = np.zeros((n_bond, 2), dtype='int')

        bond_index = 0

        atom_seq = mol.GetAtoms()
        atoms = [atom_seq[i] for i in range(n_atom)]

        for n, atom in enumerate(atoms):

            # Atom Classes
            atom_feature_matrix[n] = self.atom_tokenizer(
                self.atom_features(atom))

            start_index = atom.GetIdx()

            for bond in atom.GetBonds():
                # Is the bond pointing at the target atom
                rev = bond.GetBeginAtomIdx() != start_index

                # Bond Classes
                bond_feature_matrix[n] = self.bond_tokenizer(
                    self.bond_features(bond, flipped=rev))

                # Connectivity
                if not rev:  # Original direction
                    connectivity[bond_index, 0] = bond.GetBeginAtomIdx()
                    connectivity[bond_index, 1] = bond.GetEndAtomIdx()

                else:  # Reversed
                    connectivity[bond_index, 0] = bond.GetEndAtomIdx()
                    connectivity[bond_index, 1] = bond.GetBeginAtomIdx()

                bond_index += 1

        return {
            'n_atom': n_atom,
            'n_bond': n_bond,
            'atom': atom_feature_matrix,
            'bond': bond_feature_matrix,
            'connectivity': connectivity,
        }


class MolPreprocessor(SmilesPreprocessor):
    """ I should refactor this into a base class and separate
    SmilesPreprocessor classes. But the idea is that we only need to redefine
    the `construct_feature_matrices` method to have a working preprocessor that
    handles 3D structures.

    We'll pass an iterator of mol objects instead of SMILES strings this time,
    though.

    """

    def __init__(self, n_neighbors, cutoff, **kwargs):
        """ A preprocessor class that also returns distances between
        neighboring atoms. Adds edges for non-bonded atoms to include a maximum
        of n_neighbors around each atom """

        self.n_neighbors = n_neighbors
        self.cutoff = cutoff
        super(MolPreprocessor, self).__init__(**kwargs)


    def construct_feature_matrices(self, mol):
        """ Given an rdkit mol, return atom feature matrices, bond feature
        matrices, and connectivity matrices.

        Returns
        dict with entries
        'n_atom' : number of atoms in the molecule
        'n_bond' : number of edges (likely n_atom * n_neighbors)
        'atom' : (n_atom,) length list of atom classes
        'bond' : (n_bond,) list of bond classes. 0 for no bond
        'distance' : (n_bond,) list of bond distances
        'connectivity' : (n_bond, 2) array of source atom, target atom pairs.

        """

        n_atom = len(mol.GetAtoms())

        # n_bond is actually the number of atom-atom pairs, so this is defined
        # by the number of neighbors for each atom.
        #if there is cutoff,
        distance_matrix = Chem.Get3DDistanceMatrix(mol)

        if self.n_neighbors <= (n_atom - 1):
            n_bond = self.n_neighbors * n_atom
        else:
            # If there are fewer atoms than n_neighbors, all atoms will be
            # connected
            n_bond = distance_matrix[(distance_matrix < self.cutoff) & (distance_matrix != 0)].size

        if n_bond == 0: n_bond = 1

        # Initialize the matrices to be filled in during the following loop.
        atom_feature_matrix = np.zeros(n_atom, dtype='int')
        bond_feature_matrix = np.zeros(n_bond, dtype='int')
        bond_distance_matrix = np.zeros(n_bond, dtype=np.float32)
        connectivity = np.zeros((n_bond, 2), dtype='int')

        # Hopefully we've filtered out all problem mols by now.
        if mol is None:
            raise RuntimeError("Issue in loading mol")

        # Get a list of the atoms in the molecule.
        atom_seq = mol.GetAtoms()
        atoms = [atom_seq[i] for i in range(n_atom)]

        # Here we loop over each atom, and the inner loop iterates over each
        # neighbor of the current atom.
        bond_index = 0  # keep track of our current bond.
        for n, atom in enumerate(atoms):

            # update atom feature matrix
            atom_feature_matrix[n] = self.atom_tokenizer(
                self.atom_features(atom))

            # if n_neighbors is greater than total atoms, then each atom is a
            # neighbor.
            if (self.n_neighbors + 1) > len(mol.GetAtoms()):
                neighbor_end_index = len(mol.GetAtoms())
            else:
                neighbor_end_index = (self.n_neighbors + 1)

            distance_atom = distance_matrix[n, :]
            cutoff_end_index = distance_atom[distance_atom < self.cutoff].size

            end_index = min(neighbor_end_index, cutoff_end_index)

            # Loop over each of the nearest neighbors

            neighbor_inds = distance_matrix[n, :].argsort()[1:end_index]
            if len(neighbor_inds)==0: neighbor_inds = [n]
            for neighbor in neighbor_inds:

                # update bond feature matrix
                bond = mol.GetBondBetweenAtoms(n, int(neighbor))
                if bond is None:
                    bond_feature_matrix[bond_index] = 0
                else:
                    rev = False if bond.GetBeginAtomIdx() == n else True
                    bond_feature_matrix[bond_index] = self.bond_tokenizer(
                        self.bond_features(bond, flipped=rev))

                distance = distance_matrix[n, neighbor]
                bond_distance_matrix[bond_index] = distance

                # update connectivity matrix
                connectivity[bond_index, 0] = n
                connectivity[bond_index, 1] = neighbor

                bond_index += 1
        print(connectivity)

        return {
            'n_atom': n_atom,
            'n_bond': n_bond,
            'atom': atom_feature_matrix,
            'bond': bond_feature_matrix,
            'distance': bond_distance_matrix,
            'connectivity': connectivity,
        }


class MolBPreprocessor(MolPreprocessor):
    """
    This is a subclass of Molpreprocessor that preprocessor molecule with
    bond property target
    """
    def __init__(self, **kwargs):
        """
        A preprocessor class that also returns bond_target_matrix, besides the bond matrix
        returned by MolPreprocessor. The bond_target_matrix is then used as ref to reduce molecule
        to bond property
        """
        super(MolBPreprocessor, self).__init__(**kwargs)

    def construct_feature_matrices(self, entry):
        """
        Given an entry contining rdkit molecule, bond_index and for the target property,
        return atom
        feature matrices, bond feature matrices, distance matrices, connectivity matrices and bond
        ref matrices.

        returns
        dict with entries
        see MolPreproccessor
        'bond_index' : ref array to the bond index
        """
        mol, bond_index_array = entry

        n_atom = len(mol.GetAtoms())
        n_pro = len(bond_index_array)

        # n_bond is actually the number of atom-atom pairs, so this is defined
        # by the number of neighbors for each atom.
        #if there is cutoff,
        distance_matrix = Chem.Get3DDistanceMatrix(mol)

        if self.n_neighbors <= (n_atom - 1):
            n_bond = self.n_neighbors * n_atom
        else:
            # If there are fewer atoms than n_neighbors, all atoms will be
            # connected
            n_bond = distance_matrix[(distance_matrix < self.cutoff) & (distance_matrix != 0)].size

        if n_bond == 0: n_bond = 1

        # Initialize the matrices to be filled in during the following loop.
        atom_feature_matrix = np.zeros(n_atom, dtype='int')
        bond_feature_matrix = np.zeros(n_bond, dtype='int')
        bond_distance_matrix = np.zeros(n_bond, dtype=np.float32)
        bond_index_matrix = np.full(n_bond, -1, dtype='int')
        connectivity = np.zeros((n_bond, 2), dtype='int')

        # Hopefully we've filtered out all problem mols by now.
        if mol is None:
            raise RuntimeError("Issue in loading mol")

        # Get a list of the atoms in the molecule.
        atom_seq = mol.GetAtoms()
        atoms = [atom_seq[i] for i in range(n_atom)]

        # Here we loop over each atom, and the inner loop iterates over each
        # neighbor of the current atom.
        bond_index = 0  # keep track of our current bond.
        for n, atom in enumerate(atoms):
            # update atom feature matrix
            atom_feature_matrix[n] = self.atom_tokenizer(
                self.atom_features(atom))

            # if n_neighbors is greater than total atoms, then each atom is a
            # neighbor.
            if (self.n_neighbors + 1) > len(mol.GetAtoms()):
                neighbor_end_index = len(mol.GetAtoms())
            else:
                neighbor_end_index = (self.n_neighbors + 1)

            distance_atom = distance_matrix[n, :]
            cutoff_end_index = distance_atom[distance_atom < self.cutoff].size

            end_index = min(neighbor_end_index, cutoff_end_index)

            # Loop over each of the nearest neighbors

            neighbor_inds = distance_matrix[n, :].argsort()[1:end_index]
            if len(neighbor_inds)==0: neighbor_inds = [n]
            for neighbor in neighbor_inds:

                # update bond feature matrix
                bond = mol.GetBondBetweenAtoms(n, int(neighbor))
                if bond is None:
                    bond_feature_matrix[bond_index] = 0
                else:
                    rev = False if bond.GetBeginAtomIdx() == n else True
                    bond_feature_matrix[bond_index] = self.bond_tokenizer(
                        self.bond_features(bond, flipped=rev))
                    try:
                        bond_index_matrix[bond_index] = bond_index_array.tolist().index(bond.GetIdx())
                    except:
                        pass

                distance = distance_matrix[n, neighbor]
                bond_distance_matrix[bond_index] = distance

                # update connectivity matrix
                connectivity[bond_index, 0] = n
                connectivity[bond_index, 1] = neighbor

                bond_index += 1
        return {
            'n_atom': n_atom,
            'n_bond': n_bond,
            'n_pro': n_pro,
            'atom': atom_feature_matrix,
            'bond': bond_feature_matrix,
            'distance': bond_distance_matrix,
            'connectivity': connectivity,
            'bond_index': bond_index_matrix,
        }

class MolAPreprocessor(MolPreprocessor):
    """
    This is a subclass of Molpreprocessor that preprocessor molecule with
    bond property target
    """
    def __init__(self, **kwargs):
        """
        A preprocessor class that also returns bond_target_matrix, besides the bond matrix
        returned by MolPreprocessor. The bond_target_matrix is then used as ref to reduce molecule
        to bond property
        """
        super(MolAPreprocessor, self).__init__(**kwargs)

    def construct_feature_matrices(self, entry):
        """
        Given an entry contining rdkit molecule, bond_index and for the target property,
        return atom
        feature matrices, bond feature matrices, distance matrices, connectivity matrices and bond
        ref matrices.

        returns
        dict with entries
        see MolPreproccessor
        'bond_index' : ref array to the bond index
        """
        mol, atom_index_array = entry

        n_atom = len(mol.GetAtoms())
        n_pro = len(atom_index_array)

        # n_bond is actually the number of atom-atom pairs, so this is defined
        # by the number of neighbors for each atom.
        #if there is cutoff,
        distance_matrix = Chem.Get3DDistanceMatrix(mol)

        #if self.n_neighbors <= (n_atom - 1):
        #    n_bond = self.n_neighbors * n_atom
        #else:
            # If there are fewer atoms than n_neighbors, all atoms will be
            # connected
        n_bond = distance_matrix[(distance_matrix < self.cutoff) & (distance_matrix != 0)].size

        if n_bond == 0: n_bond = 1

        # Initialize the matrices to be filled in during the following loop.
        atom_feature_matrix = np.zeros(n_atom, dtype='int')
        bond_feature_matrix = np.zeros(n_bond, dtype='int')
        bond_distance_matrix = np.zeros(n_bond, dtype=np.float32)
        atom_index_matrix = np.full(n_atom, -1, dtype='int')
        connectivity = np.zeros((n_bond, 2), dtype='int')

        # Hopefully we've filtered out all problem mols by now.
        if mol is None:
            raise RuntimeError("Issue in loading mol")

        # Get a list of the atoms in the molecule.
        atom_seq = mol.GetAtoms()
        atoms = [atom_seq[i] for i in range(n_atom)]

        # Here we loop over each atom, and the inner loop iterates over each
        # neighbor of the current atom.
        bond_index = 0  # keep track of our current bond.
        for n, atom in enumerate(atoms):
            # update atom feature matrix
            atom_feature_matrix[n] = self.atom_tokenizer(
                self.atom_features(atom))
            try:
                atom_index_matrix[n] = atom_index_array.tolist().index(atom.GetIdx())
            except:
                pass
            # if n_neighbors is greater than total atoms, then each atom is a
            # neighbor.
            if (self.n_neighbors + 1) > len(mol.GetAtoms()):
                neighbor_end_index = len(mol.GetAtoms())
            else:
                neighbor_end_index = (self.n_neighbors + 1)

            distance_atom = distance_matrix[n, :]
            cutoff_end_index = distance_atom[distance_atom < self.cutoff].size

            end_index = min(neighbor_end_index, cutoff_end_index)

            # Loop over each of the nearest neighbors

            neighbor_inds = distance_matrix[n, :].argsort()[1:end_index]
            if len(neighbor_inds)==0: neighbor_inds = [n]
            for neighbor in neighbor_inds:

                # update bond feature matrix
                bond = mol.GetBondBetweenAtoms(n, int(neighbor))
                if bond is None:
                    bond_feature_matrix[bond_index] = 0
                else:
                    rev = False if bond.GetBeginAtomIdx() == n else True
                    bond_feature_matrix[bond_index] = self.bond_tokenizer(
                        self.bond_features(bond, flipped=rev))

                distance = distance_matrix[n, neighbor]
                bond_distance_matrix[bond_index] = distance

                # update connectivity matrix
                connectivity[bond_index, 0] = n
                connectivity[bond_index, 1] = neighbor

                bond_index += 1
        return {
            'n_atom': n_atom,
            'n_bond': n_bond,
            'n_pro': n_pro,
            'atom': atom_feature_matrix,
            'bond': bond_feature_matrix,
            'distance': bond_distance_matrix,
            'connectivity': connectivity,
            'atom_index': atom_index_matrix,
        }


# TODO: rewrite this
# class LaplacianSmilesPreprocessor(SmilesPreprocessor):
#     """ Extends the SmilesPreprocessor class to also return eigenvalues and
#     eigenvectors of the graph laplacian matrix.
#
#     Example:
#     >>> preprocessor = SmilesPreprocessor(
#     >>>     max_atoms=55, max_bonds=62, max_degree=4, explicit_hs=False)
#     >>> atom, connectivity, eigenvalues, eigenvectors = preprocessor.fit(
#             data.smiles)
#     """
#
#     def preprocess(self, smiles_iterator, train=True):
#
#         self.atom_tokenizer.train = train
#         self.bond_tokenizer.train = train
#
#         for smiles in tqdm(smiles_iterator):
#             G = self._mol_to_nx(smiles)
#             A = self._get_atom_feature_matrix(G)
#             C = self._get_connectivity_matrix(G)
#             W, V = self._get_laplacian_spectral_decomp(G)
#             yield A, C, W, V
#
#
#     def _get_laplacian_spectral_decomp(self, G):
#         """ Return the eigenvalues and eigenvectors of the graph G, padded to
#         `self.max_atoms`.
#         """
#
#         w0 = np.zeros((self.max_atoms, 1))
#         v0 = np.zeros((self.max_atoms, self.max_atoms))
#
#         w, v = eigh(nx.laplacian_matrix(G).todense())
#
#         num_atoms = len(v)
#
#         w0[:num_atoms, 0] = w
#         v0[:num_atoms, :num_atoms] = v
#
#         return w0, v0
#
#
#     def fit(self, smiles_iterator):
#         results = self._fit(smiles_iterator)
#         return {'atom': results[0],
#                 'connectivity': results[1],
#                 'w': results[2],
#                 'v': results[3]}
#
#
#     def predict(self, smiles_iterator):
#         results = self._predict(smiles_iterator)
#         return {'atom': results[0],
#                 'connectivity': results[1],
#                 'w': results[2],
#                 'v': results[3]}


def get_max_atom_bond_size(smiles_iterator, explicit_hs=True):
    """ Convienence function to get max_atoms, max_bonds for a set of input
    SMILES """

    max_atoms = 0
    max_bonds = 0
    for smiles in tqdm(smiles_iterator):
        mol = MolFromSmiles(smiles)
        if explicit_hs:
            mol = AddHs(mol)
        max_atoms = max([max_atoms, len(mol.GetAtoms())])
        max_bonds = max([max_bonds, len(mol.GetBonds())])

    return dict(max_atoms=max_atoms, max_bonds=max_bonds*2)


def canonicalize_smiles(smiles, isomeric=True, sanitize=True):
    try:
        mol = MolFromSmiles(smiles, sanitize=sanitize)
        return MolToSmiles(mol, isomericSmiles=isomeric)
    except Exception:
        pass



"""Carries out DP4 analysis of structural proposals."""

import logging
from math import prod
from pathlib import Path
import pickle
import json

import numpy as np
from scipy import stats

from .utils import scale_nmr, AnalysisData

logger = logging.getLogger(__name__)


class DP4:
    "Handles DP4 analysis"

    def __init__(self, output_folder, dp4_config):
        """Initialise DP4 Analysis object.

        Specify the save path, load relevant statistical models, set up DP4 calculators for each nucleus

        Arguments:
         output_folder(Path): path to the output folder
         dp4_config(dict): must contain keys ``stats_model`` and ``param_file``
        Explanation:
         ``dp4_config['stats_model']`` specifies the model to use.
         ``dp4_config['param_file']`` specifies the Gaussian parameters.
         Data will be saved in the output folder.
        """

        save_dir = Path(output_folder) / "dp4"
        save_dir.mkdir(exist_ok=True)
        self.save_dir = save_dir

        # define default statictical parameters
        meanC = 0.0
        meanH = 0.0
        stdevC = 2.269372270818724
        stdevH = 0.18731058105269952

        stats_model = dp4_config["stats_model"]
        param_file = dp4_config["param_file"]

        if stats_model in ("g", "m"):
            logger.info("Statistical model parameter file: %s" % param_file)
            if param_file in ("none", ""):
                logger.info("Using default statistical model")

            else:
                logger.info("Reading DP4 parameters from %s" % str(param_file))
                meanC, stdevC, meanH, stdevH = self.read_parameters(param_file)
        else:
            logger.error("Statistical model not recognised: %s" % stats_model)
            logger.info("Using default statistical model")
        self.H_probability = DP4ProbabilityCalculator(mean=meanH, stdev=stdevH)
        self.C_probability = DP4ProbabilityCalculator(mean=meanC, stdev=stdevC)

    def read_parameters(self, file):
        """Read distribution parameters from the external file
        lines of the file: comment, carbon mean errors, carbon standard deviations,
        proton mean errors, proton standard deviations
        Arguments:
        - file(str or Path): filename
        """
        with open(file, "r") as f:
            inp = f.readlines()

        Cmeans = [float(x) for x in inp[1].split(",")]
        Cstdevs = [float(x) for x in inp[2].split(",")]
        Hmeans = [float(x) for x in inp[3].split(",")]
        Hstdevs = [float(x) for x in inp[4].split(",")]
        # necessary to prevent failure at evaluation stage
        if len(Cmeans) != len(Cstdevs) or len(Hmeans) != len(Hstdevs):
            logger.critical(
                "Number of parameters does not match. Please check the file."
            )
            raise ValueError(
                "The DP4 parameters file could not be read correctly. Please check the file"
            )

        return Cmeans, Cstdevs, Hmeans, Hstdevs

    def __call__(self, mols):
        """Compute DP4 probability for the molecules
        return dictionary containing the data?
        """
        if len(mols) < 2:
            logger.warning("DP4 score requires multiple candidate structures.")
        logger.info("Starting DP4 analysis")
        dp4_dicts = DP4Data(mols, self.save_dir / "data_dic.p")
        C_data = []
        H_data = []
        keys = [
            "{0}labels",
            "{0}shifts",
            "{0}scaled",
            "{0}exp",
            "{0}errors",
            "{0}probs",
        ]
        H_keys = [i.format("H") for i in keys]
        C_keys = [i.format("C") for i in keys]
        for mol in mols:
            mol_dict = dict()

            *H_metadata, H_dp4 = self.dp4_proton(mol.H_shifts, mol.H_exp, mol.H_labels)
            H_dict = {k: v for k, v in zip(H_keys, H_metadata)}
            mol_dict.update(H_dict)
            H_data.append(H_dp4)

            *C_metadata, C_dp4 = self.dp4_carbon(mol.C_shifts, mol.C_exp, mol.C_labels)
            C_dict = {k: v for k, v in zip(C_keys, C_metadata)}
            mol_dict.update(C_dict)
            C_data.append(C_dp4)

            dp4_dicts.append(mol_dict)

        C_data = np.array(C_data)
        H_data = np.array(H_data)
        total_data = C_data * H_data

        dp4_dicts.CDP4probs = C_data / C_data.sum()
        dp4_dicts.HDP4probs = H_data / H_data.sum()
        dp4_dicts.DP4probs = total_data / total_data.sum()

        logger.info("Saving raw DP4 data")
        dp4_dicts.save()
        return dp4_dicts.output

    def dp4_proton(self, calculated, experimental, labels):
        """Generates unscaled DP4 score for protons in the molecule.
        Arguments:
         calculated: scaled calculated NMR shifts
         experimental: experimental NMR shifts
         labels: atom labels, match the SD File numbering
        returns:
         scaled calculated NMR shifts used in the analysis
         experimental NMR shifts used in the analysis
         atom labels used in the analysis
         atomic errors
         probabilities
         DP4 scores
        """
        return self._dp4(calculated, experimental, labels, self.H_probability)

    def dp4_carbon(self, calculated, experimental, labels):
        """Generates unscaled DP4 score for carbons in the molecule.
        Arguments:
         calculated: scaled calculated NMR shifts
         experimental: experimental NMR shifts
         labels: atom labels, match the SD File numbering
        returns:
         scaled calculated NMR shifts used in the analysis
         experimental NMR shifts used in the analysis
         atom labels used in the analysis
         atomic errors
         probabilities
         DP4 scores
        """
        return self._dp4(calculated, experimental, labels, self.C_probability)

    def _dp4(self, calculated: list, experimental: list, labels: list, probability):
        """Generates unscaled DP4 score for nuclei in the molecule.
        Arguments:
         calculated: scaled calculated NMR shifts
         experimental: experimental NMR shifts
         labels: atom labels, match the SD File numbering
         probability(function): returns the number
        returns:
         scaled calculated NMR shifts used in the analysis
         experimental NMR shifts used in the analysis
         atom labels used in the analysis
         atomic errors
         probabilities
         DP4 scores
        """
        # remove calculated peaks that do not match the signal
        has_exp = np.isfinite(experimental)

        new_calcs = calculated[has_exp]
        new_exps = experimental[has_exp]
        new_labs = labels[has_exp]

        new_scaled = scale_nmr(new_calcs, new_exps)
        errors = new_scaled - new_exps
        probs = probability(errors)
        # take the product of probabilities
        dp4_score = prod(probs, start=1)
        return new_labs, new_calcs, new_scaled, new_exps, errors, probs, dp4_score


class DP4ProbabilityCalculator:
    """Estimates DP4 probability for a given nucleus."""

    def __init__(self, mean, stdev):
        if isinstance(mean, list) and isinstance(stdev, list):
            # the check is redundant
            # see assert statement in DP4.read_parameters
            if len(mean) != len(stdev):
                raise ValueError(
                    "Dimensions of mean and standard deviation do not match"
                )
            if len(mean) > 1:
                self.probability = lambda error: self.multiple_gaussian(
                    error, mean, stdev
                )
            else:
                self.probability = lambda error: self.single_gaussian(
                    error, mean[0], stdev[0]
                )
        elif isinstance(mean, float) and isinstance(stdev, float):
            self.probability = lambda error: self.single_gaussian(error, mean, stdev)
        else:
            raise TypeError("Types of mean and standard deviation do not match!")

    def __call__(self, error):
        # numpy is weird about arrays out of single floats
        dp4_score = np.apply_along_axis(
            self.probability, 0, np.array(error, dtype=np.float32)
        )
        return dp4_score

    @staticmethod
    def single_gaussian(error, mean, stdev):
        z = abs((error - mean) / stdev)
        cdp4 = 2 * stats.norm.cdf(-z)

        return cdp4

    @staticmethod
    def multiple_gaussian(error, means, stdevs):
        res = 0
        for mean, stdev in zip(means, stdevs):
            res += stats.norm(mean, stdev).pdf(error)

        return res / len(means)


class DP4Data(AnalysisData):
    """Collates DP4 Analysis Data. Saves it as a pickle file and returns text summary for printing."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def output(self):
        output_dict = dict()
        output_dict["C_output"] = []
        output_dict["H_output"] = []
        output_dict["CDP4_output"] = []
        output_dict["HDP4_output"] = []
        output_dict["DP4_output"] = []

        for mol, clab, cshift, cscal, cexp, cerr in zip(
            self.mols, self.Clabels, self.Cshifts, self.Cscaled, self.Cexp, self.Cerrors
        ):
            output = f"\nAssigned C NMR shift for {mol}:"
            output += self.print_assignment(clab, cshift, cscal, cexp, cerr)
            output_dict["C_output"].append(output)

        for mol, hlab, hshift, hscal, hexp, herr in zip(
            self.mols, self.Hlabels, self.Hshifts, self.Hscaled, self.Hexp, self.Herrors
        ):
            output = f"\nAssigned H NMR shift for {mol}:"
            output += self.print_assignment(hlab, hshift, hscal, hexp, herr)
            output_dict["H_output"].append(output)
        for mol, hdp4, cdp4, dp4 in zip(
            self.mols, self.HDP4probs, self.CDP4probs, self.DP4probs
        ):
            output_dict["HDP4_output"].append(
                f"Proton DP4 probability for {mol}: {hdp4}"
            )
            output_dict["CDP4_output"].append(
                f"Carbon DP4 probability for {mol}: {cdp4}"
            )
            output_dict["DP4_output"].append(f"DP4 probability for {mol}: {dp4}")

        t_dic = [
            dict(zip(output_dict.keys(), values))
            for values in zip(*output_dict.values())
        ]
        dp4_output = "\n\n".join([mol["C_output"] + mol["H_output"] for mol in t_dic])
        dp4_output += "\n\n"
        dp4_output += "\n".join([mol["HDP4_output"] for mol in t_dic])
        dp4_output += "\n\n"
        dp4_output += "\n".join([mol["CDP4_output"] for mol in t_dic])
        dp4_output += "\n\n"
        dp4_output += "\n".join([mol["DP4_output"] for mol in t_dic])

        return dp4_output

    @staticmethod
    def print_assignment(labels, calculated, scaled, exp, error):
        """Prints table for molecule"""

        s = np.argsort(calculated)
        svalues = calculated[s]
        slabels = labels[s]
        sscaled = scaled[s]
        sexp = exp[s]
        serror = error[s]

        output = f"\nlabel, calc, corrected, exp, error"

        for lab, calc, scal, ex, er in zip(slabels, svalues, sscaled, sexp, serror):
            output += f"\n{lab:6s} {calc:6.2f} {scal:6.2f} {ex:6.2f} {er:6.2f}"
        return output



from dp5.nmr_processing.proton.integrate import find_integrals, weighted_region_centres
from dp5.nmr_processing.proton.peak_picking import gradient_peak_picking
from dp5.nmr_processing.proton.fid_corrections import spectral_processing
from dp5.nmr_processing.proton.solvent_removal import solvent_removal
from dp5.nmr_processing.proton.bic_minimisation import multiproc_BIC_minimisation
from dp5.nmr_processing.proton.simulate import simulate_regions
from dp5.nmr_processing.proton.assign import iterative_assignment



"""
The main script. Replaces PyDP4.py, orchestrates the high-level workflow.

Molecules is the container class to contain all the data. MM and DFT methods no longer modify it directly to improve maintainability.
"""

import logging

from dp5.run.data_structures import Molecules
from dp5.nmr_processing import NMRData


logger = logging.getLogger(__name__)


def runner(config):
    logger.info("Starting DP4 workflow")

    data = Molecules(config)

    if config["workflow"]["conf_search"] and not (
        config["workflow"]["restart_dft"] or config["workflow"]["calculations_complete"]
    ):

        logger.info("Conformational search requested")
        data.get_conformers()

    else:
        logger.info("No conformational search requested")

    if (
        config["workflow"]["dft_energies"]
        or config["workflow"]["dft_nmr"]
        or config["workflow"]["dft_opt"]
    ):

        logger.info("DFT calculations requested")
        data.get_dft_data()

    else:
        logger.info("No DFT calculations requested")

    if not config["workflow"]["dft_nmr"]:
        logger.info("Generating chemical shifts using a neural network")
        data.get_nn_nmr_shifts()

    # heading into legacy code area
    nmr_data = NMRData(
        config["nmr_file"],
        config["solvent"],
        config["output_folder"],
    )
    # process data first!!!!
    data.assign_nmr_spectra(nmr_data)

    # now that we have assigned it, time for DP4
    if config["workflow"]["dp5"]:
        data.dp5_analysis()
    if config["workflow"]["dp4"]:
        data.dp4_analysis()

    data.print_results()



"""
The goal is to parse configuration. 

First import config file to supersede default, 
and override I/O from command line if required.

Takes one file with line-delineated Smiles or InChIs, or several SDFiles.
Returns the final config for the run
"""

from pathlib import Path
import argparse
import os
import logging

import tomli
import json

from .logger import setup_logger

LOGLEVEL_CHOICES = tuple(level.lower() for level in logging._nameToLevel.keys())

DEFAULT_BASE_CONFIG_PATH = (
    Path(__file__).parent / "config/default_config.toml"
).resolve()


def main():

    parser = argparse.ArgumentParser(
        description="Load config and start the workflow manager."
    )

    parser.add_argument(
        "-s",
        "--structure_files",
        nargs="+",
        default=[],
        type=str,
        help="One or more SDF file for the structures to be verified by DP4. At least one\
    is required, if automatic diastereomer generation is used.",
    )

    parser.add_argument(
        "-n",
        "--nmr_file",
        nargs="+",
        help="Experimental NMR description, assigned\
    with the atom numbers from the structure file",
    )

    parser.add_argument(
        "-c",
        "--config",
        help="Load a config file",
        type=str,
        default=DEFAULT_BASE_CONFIG_PATH,
    )

    parser.add_argument(
        "-o",
        "--output",
        help="Output directory for calculations, default is current working directory.",
        default="",
    )

    parser.add_argument(
        "-i",
        "--input_type",
        help="Input file format. Default is sdf.",
        choices=["sdf", "smiles", "smarts", "inchi"],
        default="sdf",
    )

    parser.add_argument(
        "-w",
        "--workflow",
        help="Defines which steps to include in the workflow, "
        "can contain g for generate diastereomers, m for molecular mechanics conformational search, "
        "o for DFT optimization, e for DFT single-point energies, n for DFT NMR calculation, "
        "a for computational and experimental NMR data extraction, "
        "s for computational and experimental NMR data extraction and stats analysis, "
        "w for DP5 probability calculation.",
        required=False,
    )

    parser.add_argument(
        "--stereocentres",
        nargs="*",
        default=[],
        type=int,
        help="Atom indices matching input SD File for stereocentres to mutate, keep the rest intact.",
    )

    parser.add_argument("-l", "--log_filename", help="Path to log file", default="")

    parser.add_argument("--log_level", choices=LOGLEVEL_CHOICES)

    args = parser.parse_args()

    # Delay heavy imports until after CLI parsing so `pydp4 -h` returns quickly.
    from dp5.run import prepare_inputs
    from .runner import runner

    # load custom configuration
    config_path = (Path.cwd() / args.config).resolve()
    if config_path.suffix == ".toml":
        with open(config_path, "rb") as f:
            config = tomli.load(f)
    elif config_path.suffix == ".json":
        with open(config_path, "rb") as f:
            config = json.load(f)

    # commandline overrides all config files
    if args.log_level is not None:
        config["log_level"] = args.log_level
    logger = setup_logger(
        name=__package__,
        level=config["log_level"].upper(),
        filename=args.log_filename,
        propagate=True,
    )

    logger.info("Preparing configuration")
    # Override workflow flag
    if args.workflow is not None:
        config["workflow"]["cleanup"] = "c" in args.workflow
        config["workflow"]["generate"] = "g" in args.workflow
        config["workflow"]["conf_search"] = "m" in args.workflow
        config["workflow"]["dft_nmr"] = "n" in args.workflow
        config["workflow"]["dft_energies"] = "e" in args.workflow
        config["workflow"]["dft_opt"] = "o" in args.workflow
        config["workflow"]["dp4"] = "s" in args.workflow
        config["workflow"]["dp5"] = "w" in args.workflow
        config["workflow"]["assign_only"] = "a" in args.workflow

    logger.info(f"Structure cleanup required: {config['workflow']['cleanup']}")
    logger.info(f"Diastereomer generation required: {config['workflow']['generate']}")
    logger.info(f"Conformational search required: {config['workflow']['conf_search']}")
    logger.info(f"DFT geometry optimisation required: {config['workflow']['dft_opt']}")
    logger.info(
        f"DFT energy calculation required: {config['workflow']['dft_energies']}"
    )
    logger.info(f"DFT shielding calculation required: {config['workflow']['dft_nmr']}")
    logger.info(f"DP4 analysis required: {config['workflow']['dp4']}")
    logger.info(f"DP5 analysis required: {config['workflow']['dp5']}")

    # calculations_complete should set MM and DFT completion flags as True
    if config["workflow"]["calculations_complete"]:
        config["workflow"]["mm_complete"] = True
        config["dft"]["dft_complete"] = True
        config["workflow"]["dft_complete"] = True

    # reads command line argument if supplied, else reads config
    if args.structure_files:
        config["structure"] = args.structure_files
        config["input_type"] = args.input_type
        config["stereocentres"] = args.stereocentres
        logger.debug(
            f"Read structures {', '.join(config['structure'])} from command line"
        )
    elif config["structure"]:
        logger.debug(
            f"Read structures {', '.join(config['structure'])} from config file"
        )
    else:
        logger.critical("No structures specified")
        raise ValueError("No structures specified")

    logger.info(f"Structure input files: {', '.join(config['structure'])}")

    if args.nmr_file:
        logger.debug(f"Read NMR File {args.nmr_file} from command line")
        config["nmr_file"] = args.nmr_file
    elif config["nmr_file"]:
        logger.debug(f"Read NMR File {config['nmr_file']} from config file")
    else:
        logger.critical("No NMR data specified")
        raise ValueError("No NMR data specified")

    # set up TMS constants
    with open((Path(__file__).parent / "dft" / "TMSdata").resolve()) as file:
        _params_found = False
        _solvent = config["dft"]["solvent"] if config["dft"]["solvent"] else "none"
        for line in file:
            line = line.strip()
            if line:
                functional, basis_set, solvent, tms_c, tms_h = line.split()
                if (
                    config["dft"]["n_functional"] == functional
                    and config["dft"]["n_basis_set"] == basis_set
                    and _solvent == solvent
                ):
                    _params_found = True

                    config["dft"]["c13_tms"] = float(tms_c)
                    config["dft"]["h1_tms"] = float(tms_h)

                    break

    if not _params_found:
        logger.warning(
            "No reference shielding found for the conditions, using default values!"
        )
        functional, basis_set, solvent = ("b3lyp", "6-31G**", "none")

    logger.info("Read shielding parameters for: ")
    logger.info(
        "NMR DFT functional: %s, basis set: %s, solvent: %s",
        functional,
        basis_set,
        solvent,
    )

    logger.info(f"13C reference shielding: {config['dft']['c13_tms']:.1f} ppm")
    logger.info(f"1H reference shielding: {config['dft']['h1_tms']:.2f} ppm")

    if args.output:
        config["output_folder"] = args.output
    config["output_folder"] = (Path.cwd() / config["output_folder"]).resolve()

    config["dft"]["solvent"] = config["solvent"]

    config["structure"] = prepare_inputs(
        config["structure"],
        config["input_type"],
        config["stereocentres"],
        config["workflow"],
    )

    logger.info(f"Final structure input files:{config['structure']}")

    logger.info(f"NMR input paths:{config['nmr_file']}")

    with open(config["output_folder"] / "config.json", "w") as f:
        cfg = config.copy()
        cfg["output_folder"] = str(cfg["output_folder"])
        json.dump(cfg, f, indent=4)

    logger.info("Configuration saved to %s" % str(config["output_folder"]))

    runner(config)

    logger.info("Program terminated normally")


if __name__ == "__main__":
    main()
